# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------


# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.ActiveKnowledgeBase import ActiveKnowledgeBase
from pylstar.Letter import Letter, EmptyLetter
from pylstar.Word import Word


@PylstarLogger
class FakeActiveKnowledgeBase(ActiveKnowledgeBase):
    """An implementation of an active knowledge base that relies
    on a preseted automata to answer queries.

    
    >>> from pylstar.automata.State import State
    >>> from pylstar.automata.Transition import Transition
    >>> from pylstar.automata.Automata import Automata
    >>> from pylstar.Letter import Letter
    >>> from pylstar.Word import Word
    >>> from pylstar.OutputQuery import OutputQuery
    >>> from pylstar.FakeActiveKnowledgeBase import FakeActiveKnowledgeBase
    >>> l_a = Letter('a')
    >>> l_b = Letter('b')
    >>> l_c = Letter('c')
    >>> l_1 = Letter(1)
    >>> l_2 = Letter(2)
    >>> l_3 = Letter(3)
    >>> s0 = State("S0")
    >>> s1 = State("S1")
    >>> t1 = Transition("t1", output_state=s0, input_letter=l_a, output_letter=l_1)
    >>> t2 = Transition("t2", output_state=s1, input_letter=l_b, output_letter=l_2)
    >>> t3 = Transition("t3", output_state=s0, input_letter=l_c, output_letter=l_3)
    >>> s0.transitions = [t1, t2, t3]
    >>> t4 = Transition("t4", output_state=s1, input_letter=l_a, output_letter=l_1)
    >>> t5 = Transition("t5", output_state=s1, input_letter=l_b, output_letter=l_2)
    >>> t6 = Transition("t6", output_state=s0, input_letter=l_c, output_letter=l_3)
    >>> s1.transitions = [t4, t5, t6]
    >>> automata = Automata(s0)
    >>> kbase = FakeActiveKnowledgeBase(automata)
    >>> w1 = Word([l_a])
    >>> o1 = OutputQuery(w1)    
    >>> w2 = Word([l_b])
    >>> o2 = OutputQuery(w2)    
    >>> w3 = Word([l_a, l_b, l_b, l_b, l_b, l_c, l_c, l_a])
    >>> o3 = OutputQuery(w3)    
    >>> kbase.resolve_query(o1)
    >>> print(o1.output_word)
    [Letter(1)]
    >>> kbase.resolve_query(o2)
    >>> print(o2.output_word)
    [Letter(2)]
    >>> kbase.resolve_query(o3)
    >>> print(o3.output_word)
    [Letter(1), Letter(2), Letter(2), Letter(2), Letter(2), Letter(3), Letter(3), Letter(1)]

    """

    def __init__(self, automata):
        super(FakeActiveKnowledgeBase, self).__init__()
        self.automata = automata

    def start_target(self):
        self._logger.debug("Starting the fake target")

    def stop_target(self):
        self._logger.debug("Stoping the fake target")        

    def submit_word(self, word):
        self._logger.debug("Submiting word '{}' to the fake target".format(word))

        if self.automata is None:
            raise Exception("Automata cannot be None")
        
        current_state = self.automata.initial_state
        output_letters = []
        for letter in word.letters:
            try:
                (current_state, output_letter) = self._next_state(current_state, letter)
            except Exception:
                output_letter = EmptyLetter()
            
            output_letters.append(output_letter)

        output_word = Word(output_letters)
        return output_word

    def _next_state(self, current_state, letter):
        if current_state is None:
            raise Exception("Current state cannot be None")
        if letter is None:
            raise Exception("Letter cannot be None")

        if letter == EmptyLetter:
            return (current_state, EmptyLetter())

        for transition in current_state.transitions:
            if transition.input_letter == letter:
                return (transition.output_state, transition.output_letter)
        raise Exception("State '{}' accepts no transition triggered by letter '{}'".format(current_state, letter))
                
            

        
            

        
        

        

    



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import collections
import itertools
    
# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.Letter import EmptyLetter
from pylstar.Word import Word
from pylstar.OutputQuery import OutputQuery
from pylstar.automata.State import State
from pylstar.automata.Automata import Automata
from pylstar.automata.Transition import Transition



@PylstarLogger
class ObservationTable(object):
    """An observation table consists in a two-dimensional table,
    with rows indexed by prefixes and columns indexed by suffixes.
    Prefixes are divided in two parts, short (S) and long prefixes (SA).

    The content of the observation table is stored in the internal variable 'ot_content'
    such that :
    - O : the set of output letters
    - D : the set of input letters a.k.a Distinguising Set
    - S : the set of short prefixes
    - SA : the set of long prefixes    
    - ot_content : D x {S,SA} -> O
    
    """
    
    def __init__(self, input_letters, knowledge_base):
        self.input_letters = input_letters
        self.knowledge_base = knowledge_base

        self.initialized = False
        self.D = []
        self.S = []
        self.SA = []
        self.ot_content = dict()
        
    def initialize(self):        
        self._logger.debug("Initialization of the observation table""")
        
        if self.initialized:
            raise Exception("Observation table is already initialized")

        self.initialized = True

        self.D = []
        self.S = []
        self.SA = []
        self.ot_content = dict()

        # creates a word for each input letter and register it in D
        for letter in self.input_letters:
            self.__add_word_in_D(Word([letter]))

        # creates a word that contains an EmptyLetter and registers it in S
        self.__add_word_in_S(Word([EmptyLetter()]))
    
        #for letter in self.input_letters:
        #     self.__add_word_in_SA(Word([letter]))

    def serialize(self):
        """This method returns an str representation of the current state of observation table"""
        result = []

        result.append(str(self.D))
        result.append(str(self.S))
        result.append(str(self.SA))
        result.append(str(self.ot_content))
        
        return '\n\n\n'.join(result)
        

    def find_inconsistency(self):
        """This method returns the inconsistency found in the observation table.
        Returns None if the table is consistent.

        A table is said consistent if every equivalent pair of rows in S remains equivalent
        in SA after appending any symbol of the input vocabulary.

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> from pylstar.Letter import Letter, EmptyLetter
        >>> from pylstar.Word import Word        
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable(input_letters = [], knowledge_base = kbase)
        >>> l_lambda = EmptyLetter()
        >>> l_a = Letter('a')
        >>> l_b = Letter('b')
        >>> ot.input_letters = [l_a, l_b]
        >>> l_0 = Letter(0)
        >>> l_1 = Letter(1)
        >>> w_lambda = Word([l_lambda])
        >>> w_a = Word([l_a])
        >>> w_b = Word([l_b])
        >>> w_aa = Word([l_a, l_a])
        >>> w_ab = Word([l_a, l_b])
        >>> w_ba = Word([l_b, l_a])
        >>> w_bb = Word([l_b, l_b])
        >>> w_0 = Word([l_0])
        >>> w_1 = Word([l_1])        
        >>> ot.D = [w_a, w_b]
        >>> ot.S = [w_lambda, w_a, w_b]
        >>> ot.SA = [w_aa, w_ab, w_ba, w_bb]
        >>> ot.ot_content[w_a] = dict()
        >>> ot.ot_content[w_a][w_lambda] = w_0
        >>> ot.ot_content[w_a][w_a] = w_0
        >>> ot.ot_content[w_a][w_b] = w_1
        >>> ot.ot_content[w_a][w_aa] = w_1
        >>> ot.ot_content[w_a][w_ab] = w_1
        >>> ot.ot_content[w_a][w_ba] = w_0
        >>> ot.ot_content[w_a][w_bb] = w_0
        >>> ot.ot_content[w_b] = dict()
        >>> ot.ot_content[w_b][w_lambda] = w_0
        >>> ot.ot_content[w_b][w_a] = w_0
        >>> ot.ot_content[w_b][w_b] = w_1
        >>> ot.ot_content[w_b][w_aa] = w_1
        >>> ot.ot_content[w_b][w_ab] = w_1
        >>> ot.ot_content[w_b][w_ba] = w_0
        >>> ot.ot_content[w_b][w_bb] = w_0
        >>> print(ot) #doctest: +NORMALIZE_WHITESPACE
                                   | [Letter('a')] | [Letter('b')]
        -------------------------- | ------------- | -------------
        [EmptyLetter]              | [Letter(0)]   | [Letter(0)]  
        [Letter('a')]              | [Letter(0)]   | [Letter(0)]  
        [Letter('b')]              | [Letter(1)]   | [Letter(1)]  
        ~~~                        | ~~~           | ~~~          
        [Letter('a'), Letter('a')] | [Letter(1)]   | [Letter(1)]  
        [Letter('a'), Letter('b')] | [Letter(1)]   | [Letter(1)]  
        [Letter('b'), Letter('a')] | [Letter(0)]   | [Letter(0)]  
        [Letter('b'), Letter('b')] | [Letter(0)]   | [Letter(0)]  
        -------------------------- | ------------- | -------------
        >>> print(len(ot.find_inconsistency()))
        2
        >>> w_aaa = Word([l_a, l_a, l_a])
        >>> w_aab = Word([l_a, l_a, l_b])
        >>> ot.S.append(w_aa)
        >>> ot.SA.remove(w_aa)
        >>> ot.SA.extend([w_aaa, w_aab])
        >>> ot.ot_content[w_a][w_aa] = w_0
        >>> ot.ot_content[w_b][w_aa] = w_0
        >>> ot.ot_content[w_a][w_aaa] = w_1
        >>> ot.ot_content[w_b][w_aaa] = w_0
        >>> ot.ot_content[w_a][w_aab] = w_1 
        >>> ot.ot_content[w_b][w_aab] = w_1               
        >>> print(ot) #doctest: +NORMALIZE_WHITESPACE
                                                | [Letter('a')] | [Letter('b')]
        --------------------------------------- | ------------- | -------------
        [EmptyLetter]                           | [Letter(0)]   | [Letter(0)]  
        [Letter('a')]                           | [Letter(0)]   | [Letter(0)]  
        [Letter('b')]                           | [Letter(1)]   | [Letter(1)]  
        [Letter('a'), Letter('a')]              | [Letter(0)]   | [Letter(0)]  
        ~~~                                     | ~~~           | ~~~          
        [Letter('a'), Letter('b')]              | [Letter(1)]   | [Letter(1)]  
        [Letter('b'), Letter('a')]              | [Letter(0)]   | [Letter(0)]  
        [Letter('b'), Letter('b')]              | [Letter(0)]   | [Letter(0)]  
        [Letter('a'), Letter('a'), Letter('a')] | [Letter(1)]   | [Letter(0)]  
        [Letter('a'), Letter('a'), Letter('b')] | [Letter(1)]   | [Letter(1)]  
        --------------------------------------- | ------------- | -------------

        """
        self._logger.debug("Computes if the observation table is consistent.")

        # find all rows in S
        rows_in_S = {s: self.__get_row(s) for s in self.S}

        # identify all equivalent rows in S
        S_with_same_rows = collections.defaultdict(list)
        for word_in_s, row_s in rows_in_S.items():
            S_with_same_rows[','.join([str(w) for w in row_s])].append(word_in_s)

        # check all equivalent in S are also equivalent in SA for each
        for row_in_S, eq_words_in_S in S_with_same_rows.items():
            if len(eq_words_in_S) > 1:
                for pair_eq_words_in_S in itertools.combinations(eq_words_in_S, 2):
                    inconsistency = self.__is_prefixes_equivalent(pair_eq_words_in_S)
                    if inconsistency is not None:
                        suffix, inconsistency_detail = inconsistency
                        return ((pair_eq_words_in_S, suffix), inconsistency_detail)
        return None

    def __is_prefixes_equivalent(self, eq_words_in_S):
        """This method checks that the specified prefixes are equivalent.

        It returns None, if all the prefixes share the same row value given any
        input letters it get suffixed with
        Returns the letter that makes the eq_words_in_S not equivalent

        """

        if len(eq_words_in_S) < 2:
            raise Exception("At least two words must be provided")

        self._logger.debug("Checking if words '{}' are equivalents".format(','.join([str(s) for s in eq_words_in_S])))

        for input_letter in self.input_letters:
            initial_suffixed_word = Word(eq_words_in_S[0].letters + [input_letter])
            for eq_word_in_S in eq_words_in_S[1:]:                
                eq_suffixed_word = Word(eq_word_in_S.letters + [input_letter])
                for word_in_D in self.D:
                    self._logger.debug(type(word_in_D))
                    if self.ot_content[word_in_D][eq_suffixed_word] !=  self.ot_content[word_in_D][initial_suffixed_word]:
                        return (input_letter, word_in_D)
                
        return None

    def add_counterexample(self, input_word, output_word):
        """This method register the specified counterexample in the observation table.

        In details, it insert all the prefixes of the counterexample to the upper part of the table.
        It also extend SA accordingly.

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.FakeActiveKnowledgeBase import FakeActiveKnowledgeBase
        >>> from pylstar.Letter import Letter, EmptyLetter
        >>> from pylstar.Word import Word
        >>> from pylstar.automata.State import State
        >>> from pylstar.automata.Transition import Transition
        >>> from pylstar.automata.Automata import Automata
        >>> l_lambda = EmptyLetter()
        >>> l_a = Letter('a')
        >>> l_b = Letter('b')
        >>> l_0 = Letter(0)
        >>> l_1 = Letter(1)
        >>> s0 = State("S0")
        >>> s1 = State("S1")
        >>> s2 = State("S2")
        >>> s3 = State("S3")
        >>> t1 = Transition("T1", s3, l_a, l_0)
        >>> t2 = Transition("T2", s1, l_b, l_0)        
        >>> s0.transitions = [t1, t2]
        >>> t3 = Transition("T3", s0, l_a, l_1)
        >>> t4 = Transition("T4", s2, l_b, l_1)        
        >>> s1.transitions = [t3, t4]
        >>> t5 = Transition("T5", s3, l_a, l_0)
        >>> t6 = Transition("T6", s0, l_b, l_0)        
        >>> s2.transitions = [t5, t6]        
        >>> t7 = Transition("T7", s3, l_a, l_1)
        >>> t8 = Transition("T8", s3, l_b, l_1)        
        >>> s3.transitions = [t7, t8]
        >>> automata = Automata(s0)
        >>> kbase = FakeActiveKnowledgeBase(automata)
        >>> ot = ObservationTable(input_letters = [l_a, l_b], knowledge_base = kbase)
        >>> ot.initialize()
        >>> print(ot)
                      | [Letter('a')] | [Letter('b')]
        ------------- | ------------- | -------------
        [EmptyLetter] | Letter(0)     | Letter(0)    
        ~~~           | ~~~           | ~~~          
        [Letter('a')] | Letter(1)     | Letter(1)    
        [Letter('b')] | Letter(1)     | Letter(1)    
        ------------- | ------------- | -------------
        >>> print(ot.is_closed())
        False
        >>> ot.close_table()
        >>> print(ot)
                                   | [Letter('a')] | [Letter('b')]
        -------------------------- | ------------- | -------------
        [EmptyLetter]              | Letter(0)     | Letter(0)    
        [Letter('a')]              | Letter(1)     | Letter(1)    
        ~~~                        | ~~~           | ~~~          
        [Letter('b')]              | Letter(1)     | Letter(1)    
        [Letter('a'), Letter('a')] | Letter(1)     | Letter(1)    
        [Letter('a'), Letter('b')] | Letter(1)     | Letter(1)    
        -------------------------- | ------------- | -------------
        >>> counter_input_word = Word([l_b, l_b, l_b])
        >>> counter_output_word = Word([l_0, l_1, l_0])
        >>> ot.add_counterexample(counter_input_word, counter_output_word)
        >>> print(ot)
                                                             | [Letter('a')] | [Letter('b')]
        ---------------------------------------------------- | ------------- | -------------
        [EmptyLetter]                                        | Letter(0)     | Letter(0)    
        [Letter('a')]                                        | Letter(1)     | Letter(1)    
        [Letter('b')]                                        | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('b')]                           | Letter(0)     | Letter(0)    
        [Letter('b'), Letter('b'), Letter('b')]              | Letter(0)     | Letter(0)    
        ~~~                                                  | ~~~           | ~~~          
        [Letter('a'), Letter('a')]                           | Letter(1)     | Letter(1)    
        [Letter('a'), Letter('b')]                           | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('a')]                           | Letter(0)     | Letter(0)    
        [Letter('b'), Letter('b'), Letter('a')]              | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('b'), Letter('b'), Letter('a')] | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('b'), Letter('b'), Letter('b')] | Letter(1)     | Letter(1)    
        ---------------------------------------------------- | ------------- | -------------
        
        """

        
        if input_word is None or len(input_word) == 0:
            raise Exception("Input word cannot be None or empty")
        if output_word is None or len(output_word) == 0:
            raise Exception("Output word cannot be None or empty")
        if len(input_word) != len(output_word):
            raise Exception("Output word must have the same length then input word")

        for len_prefix in range(1, len(input_word)+1):
            prefix_input = Word(input_word.letters[:len_prefix])
            if prefix_input not in self.S:
                if prefix_input in self.SA:
                    self.remove_row(prefix_input)

                self.__add_word_in_S(prefix_input)

    def remove_row(self, row):
        if row is None:
            raise Exception("Row cannot be None")
        if row in self.S:
            self.S.remove(row)
        if row in self.SA:
            self.SA.remove(row)

        for word_in_D in self.D:
            cel = self.ot_content[word_in_D]
            try:
                cel[row] = None
            except: pass
            

        

    def make_consistent(self,inconsistency):
        """This method makes consistent the observation table.

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.FakeActiveKnowledgeBase import FakeActiveKnowledgeBase
        >>> from pylstar.Letter import Letter, EmptyLetter
        >>> from pylstar.Word import Word
        >>> from pylstar.automata.State import State
        >>> from pylstar.automata.Transition import Transition
        >>> from pylstar.automata.Automata import Automata
        >>> l_lambda = EmptyLetter()
        >>> l_a = Letter('a')
        >>> l_b = Letter('b')
        >>> l_0 = Letter(0)
        >>> l_1 = Letter(1)
        >>> s0 = State("S0")
        >>> s1 = State("S1")
        >>> s2 = State("S2")
        >>> s3 = State("S3")
        >>> t1 = Transition("T1", s3, l_a, l_0)
        >>> t2 = Transition("T2", s1, l_b, l_0)        
        >>> s0.transitions = [t1, t2]
        >>> t3 = Transition("T3", s0, l_a, l_1)
        >>> t4 = Transition("T4", s2, l_b, l_1)        
        >>> s1.transitions = [t3, t4]
        >>> t5 = Transition("T5", s3, l_a, l_0)
        >>> t6 = Transition("T6", s0, l_b, l_0)        
        >>> s2.transitions = [t5, t6]        
        >>> t7 = Transition("T7", s3, l_a, l_1)
        >>> t8 = Transition("T8", s3, l_b, l_1)        
        >>> s3.transitions = [t7, t8]
        >>> automata = Automata(s0)
        >>> kbase = FakeActiveKnowledgeBase(automata)
        >>> ot = ObservationTable(input_letters = [l_a, l_b], knowledge_base = kbase)
        >>> ot.initialize()
        >>> print(ot)
                      | [Letter('a')] | [Letter('b')]
        ------------- | ------------- | -------------
        [EmptyLetter] | Letter(0)     | Letter(0)    
        ~~~           | ~~~           | ~~~          
        [Letter('a')] | Letter(1)     | Letter(1)    
        [Letter('b')] | Letter(1)     | Letter(1)    
        ------------- | ------------- | -------------
        >>> print(ot.is_closed())
        False
        >>> ot.close_table()
        >>> print(ot)
                                   | [Letter('a')] | [Letter('b')]
        -------------------------- | ------------- | -------------
        [EmptyLetter]              | Letter(0)     | Letter(0)    
        [Letter('a')]              | Letter(1)     | Letter(1)    
        ~~~                        | ~~~           | ~~~          
        [Letter('b')]              | Letter(1)     | Letter(1)    
        [Letter('a'), Letter('a')] | Letter(1)     | Letter(1)    
        [Letter('a'), Letter('b')] | Letter(1)     | Letter(1)    
        -------------------------- | ------------- | -------------
        >>> counter_input_word = Word([l_b, l_b, l_b])
        >>> counter_output_word = Word([l_0, l_1, l_0])
        >>> ot.add_counterexample(counter_input_word, counter_output_word)
        >>> print(ot)
                                                             | [Letter('a')] | [Letter('b')]
        ---------------------------------------------------- | ------------- | -------------
        [EmptyLetter]                                        | Letter(0)     | Letter(0)    
        [Letter('a')]                                        | Letter(1)     | Letter(1)    
        [Letter('b')]                                        | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('b')]                           | Letter(0)     | Letter(0)    
        [Letter('b'), Letter('b'), Letter('b')]              | Letter(0)     | Letter(0)    
        ~~~                                                  | ~~~           | ~~~          
        [Letter('a'), Letter('a')]                           | Letter(1)     | Letter(1)    
        [Letter('a'), Letter('b')]                           | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('a')]                           | Letter(0)     | Letter(0)    
        [Letter('b'), Letter('b'), Letter('a')]              | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('b'), Letter('b'), Letter('a')] | Letter(1)     | Letter(1)    
        [Letter('b'), Letter('b'), Letter('b'), Letter('b')] | Letter(1)     | Letter(1)    
        ---------------------------------------------------- | ------------- | -------------
        >>> inconsistency = ot.find_inconsistency()
        >>> ot.make_consistent(inconsistency)
        >>> inconsistency = ot.find_inconsistency()
        >>> if inconsistency is not None: ot.make_consistent(inconsistency)
        >>> print(ot.find_inconsistency())
        None
        >>> print(ot.is_closed())
        True
        >>> automata = ot.build_hypothesis()
        >>> print(automata.build_dot_code())
        digraph "Automata" {
        "0" [shape=doubleoctagon, style=filled, fillcolor=white, URL="0"];
        "2" [shape=ellipse, style=filled, fillcolor=white, URL="2"];
        "3" [shape=ellipse, style=filled, fillcolor=white, URL="3"];
        "1" [shape=ellipse, style=filled, fillcolor=white, URL="1"];
        "0" -> "1" [fontsize=5, label="a / 0", URL="t0"];
        "0" -> "2" [fontsize=5, label="b / 0", URL="t1"];
        "2" -> "0" [fontsize=5, label="a / 1", URL="t4"];
        "2" -> "3" [fontsize=5, label="b / 1", URL="t5"];
        "3" -> "1" [fontsize=5, label="a / 0", URL="t6"];
        "3" -> "0" [fontsize=5, label="b / 0", URL="t7"];
        "1" -> "1" [fontsize=5, label="a / 1", URL="t2"];
        "1" -> "1" [fontsize=5, label="b / 1", URL="t3"];
        }
                


        """

        if inconsistency is None:
            raise Exception("Inconsistency cannot be None")
        # verify both words of inconsistency share the same last letter        
        suffix = inconsistency[0][1]
        inconsistent_suffix = inconsistency[1]

        new_col_word = Word([suffix] + inconsistent_suffix.letters)
        self.__add_word_in_D(new_col_word)
        

    def is_closed(self):
        """This method returns True if the observation table is closed.

        The Observation table is said closed if all the rows in SA also exist in S.

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> from pylstar.Letter import Letter, EmptyLetter
        >>> from pylstar.Word import Word        
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable(input_letters = [], knowledge_base = kbase)
        >>> ot.initialize()
        >>> l_lambda = EmptyLetter()
        >>> l_a = Letter('a')
        >>> l_b = Letter('b')
        >>> l_0 = Letter(0)
        >>> l_1 = Letter(1)
        >>> w_lambda = Word([l_lambda])
        >>> w_a = Word([l_a])
        >>> w_b = Word([l_b])
        >>> w_aa = Word([l_a, l_a])
        >>> w_ab = Word([l_a, l_b])
        >>> w_0 = Word([l_0])
        >>> w_1 = Word([l_1])        
        >>> ot.D = [w_lambda, w_a]
        >>> ot.S = [w_lambda, w_a]
        >>> ot.SA = [w_b, w_aa, w_ab]
        >>> ot.ot_content[w_lambda] = dict()
        >>> ot.ot_content[w_lambda][w_lambda] = w_0
        >>> ot.ot_content[w_lambda][w_a] = w_1
        >>> ot.ot_content[w_lambda][w_b] = w_1
        >>> ot.ot_content[w_lambda][w_aa] = w_0
        >>> ot.ot_content[w_lambda][w_ab] = w_1        
        >>> ot.ot_content[w_a] = dict()
        >>> ot.ot_content[w_a][w_lambda] = w_1
        >>> ot.ot_content[w_a][w_a] = w_0
        >>> ot.ot_content[w_a][w_b] = w_0
        >>> ot.ot_content[w_a][w_aa] = w_1
        >>> ot.ot_content[w_a][w_ab] = w_1
        >>> print(ot)
                                   | [EmptyLetter] | [Letter('a')]
        -------------------------- | ------------- | -------------
        [EmptyLetter]              | [Letter(0)]   | [Letter(1)]  
        [Letter('a')]              | [Letter(1)]   | [Letter(0)]  
        ~~~                        | ~~~           | ~~~          
        [Letter('b')]              | [Letter(1)]   | [Letter(0)]  
        [Letter('a'), Letter('a')] | [Letter(0)]   | [Letter(1)]  
        [Letter('a'), Letter('b')] | [Letter(1)]   | [Letter(1)]  
        -------------------------- | ------------- | -------------
        >>> ot.is_closed()
        False
        >>> ot.S.append(w_ab)
        >>> ot.SA.remove(w_ab)
        >>> w_aba = Word([l_a, l_b, l_a])
        >>> w_abb = Word([l_a, l_b, l_b])        
        >>> ot.SA.extend([w_aba, w_abb])
        >>> ot.ot_content[w_lambda][w_aba] = w_0
        >>> ot.ot_content[w_lambda][w_abb] = w_1
        >>> ot.ot_content[w_a][w_aba] = w_1
        >>> ot.ot_content[w_a][w_abb] = w_0
        >>> print(ot)
                                                | [EmptyLetter] | [Letter('a')]
        --------------------------------------- | ------------- | -------------
        [EmptyLetter]                           | [Letter(0)]   | [Letter(1)]  
        [Letter('a')]                           | [Letter(1)]   | [Letter(0)]  
        [Letter('a'), Letter('b')]              | [Letter(1)]   | [Letter(1)]  
        ~~~                                     | ~~~           | ~~~          
        [Letter('b')]                           | [Letter(1)]   | [Letter(0)]  
        [Letter('a'), Letter('a')]              | [Letter(0)]   | [Letter(1)]  
        [Letter('a'), Letter('b'), Letter('a')] | [Letter(0)]   | [Letter(1)]  
        [Letter('a'), Letter('b'), Letter('b')] | [Letter(1)]   | [Letter(0)]  
        --------------------------------------- | ------------- | -------------
        >>> ot.is_closed()
        True
        
        """
        self._logger.debug("Computes if the observation table is closed")

        for sa in self.SA:
            row_sa = self.__get_row(sa)
            found = False
            for s in self.S:
                if self.__get_row(s) == row_sa:
                    found = True
                    break
            if not found :
                return False
        return True

    def close_table(self):
        """This method closes the observation table.

        To close the observation table, it moves every row from SA to S that
        have no equivalent in S

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.FakeActiveKnowledgeBase import FakeActiveKnowledgeBase
        >>> from pylstar.Letter import Letter, EmptyLetter
        >>> from pylstar.Word import Word
        >>> from pylstar.automata.State import State
        >>> from pylstar.automata.Transition import Transition
        >>> from pylstar.automata.Automata import Automata
        >>> l_lambda = EmptyLetter()
        >>> l_a = Letter('a')
        >>> l_b = Letter('b')
        >>> l_1 = Letter(1)
        >>> l_2 = Letter(2)
        >>> l_3 = Letter(3)
        >>> s0 = State("S0")
        >>> s1 = State("S1")
        >>> t1 = Transition("T1", s0, l_a, l_1)
        >>> t2 = Transition("T2", s1, l_b, l_2)
        >>> t3 = Transition("T3", s1, l_a, l_1)
        >>> t4 = Transition("T4", s0, l_b, l_3)
        >>> s0.transitions = [t1, t2]
        >>> s1.transitions = [t3, t4]
        >>> automata = Automata(s0)
        >>> kbase = FakeActiveKnowledgeBase(automata)
        >>> ot = ObservationTable(input_letters = [l_a, l_b], knowledge_base = kbase)
        >>> ot.initialize()
        >>> print(ot)
                      | [Letter('a')] | [Letter('b')]
        ------------- | ------------- | -------------
        [EmptyLetter] | Letter(1)     | Letter(2)    
        ~~~           | ~~~           | ~~~          
        [Letter('a')] | Letter(1)     | Letter(2)    
        [Letter('b')] | Letter(1)     | Letter(3)    
        ------------- | ------------- | -------------
        >>> ot.is_closed()
        False
        >>> ot.close_table()
        >>> print(ot)
                                   | [Letter('a')] | [Letter('b')]
        -------------------------- | ------------- | -------------
        [EmptyLetter]              | Letter(1)     | Letter(2)    
        [Letter('b')]              | Letter(1)     | Letter(3)    
        ~~~                        | ~~~           | ~~~          
        [Letter('a')]              | Letter(1)     | Letter(2)    
        [Letter('b'), Letter('a')] | Letter(1)     | Letter(3)    
        [Letter('b'), Letter('b')] | Letter(1)     | Letter(2)    
        -------------------------- | ------------- | -------------

        >>> ot.is_closed()
        True

        
        """

        for word_in_sa in self.SA:
            row_sa = self.__get_row(word_in_sa)
            found = False
            for word_in_s in self.S:
                row_s = self.__get_row(word_in_s)
                if row_s == row_sa:
                    found = True
                    break
            if not found:
                self._logger.debug("Row attached to SA '{}' ({}) could not be found in S".format(word_in_sa, row_sa))
                self.SA.remove(word_in_sa)
                self.__add_word_in_S(word_in_sa)
        
        self._logger.debug("Closing the observation table.")

    def __get_row(self, row_name):
        """This method returns the specified row

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable([], kbase)
        >>> ot.initialize()
        >>> w_a = Word([Letter('a')])
        >>> w_b = Word([Letter('b')])
        >>> w_c = Word([Letter('c')])
        >>> w_aa = Word([Letter('aa')])
        >>> w_ba = Word([Letter('ba')])
        >>> w_1 = Word([Letter('1')])
        >>> w_2 = Word([Letter('2')])
        >>> w_3 = Word([Letter('3')])                
        >>> ot.D.extend([w_a, w_b, w_c])
        >>> ot.S.extend([w_a, w_b])
        >>> ot.SA.extend([w_aa, w_ba])        
        >>> ot.ot_content[w_a] = dict()
        >>> ot.ot_content[w_a][w_a] = w_1
        >>> ot.ot_content[w_a][w_b] = w_1
        >>> ot.ot_content[w_a][w_aa] = w_3
        >>> ot.ot_content[w_a][w_ba] = w_3
        >>> ot.ot_content[w_b] = dict()
        >>> ot.ot_content[w_b][w_a] = w_2
        >>> ot.ot_content[w_b][w_b] = w_2
        >>> ot.ot_content[w_b][w_aa] = w_2
        >>> ot.ot_content[w_b][w_ba] = w_2
        >>> ot.ot_content[w_c] = dict()
        >>> ot.ot_content[w_c][w_a] = w_3
        >>> ot.ot_content[w_c][w_b] = w_3
        >>> ot.ot_content[w_c][w_aa] = w_1
        >>> ot.ot_content[w_c][w_ba] = w_1
        >>> print(', '.join([str(w) for w in ot._ObservationTable__get_row(w_a)]))
        [Letter('1')], [Letter('2')], [Letter('3')]
        >>> print(', '.join([str(w) for w in ot._ObservationTable__get_row(w_b)]))
        [Letter('1')], [Letter('2')], [Letter('3')]
        >>> print(', '.join([str(w) for w in ot._ObservationTable__get_row(w_aa)]))
        [Letter('3')], [Letter('2')], [Letter('1')]
        >>> print(', '.join([str(w) for w in ot._ObservationTable__get_row(w_ba)]))
        [Letter('3')], [Letter('2')], [Letter('1')]


"""
        
        if row_name is None:
            raise Exception("Row_name cannot be None")

        row = []

        for word_in_D in self.D:
            cel = self.ot_content[word_in_D]
            
            for r,v in cel.items():
                if r == row_name:
                    row.append(v)
        return row
        
        

    def __add_word_in_D(self, word):
        """

        A 'None' word cannot be inserted in D

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable(input_letters = [], knowledge_base = kbase)
        >>> ot._ObservationTable__add_word_in_D(None)
        Traceback (most recent call last):
        ...
        Exception: Word cannot be None


        A word cannot be inserted twice in D

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable(input_letters = [], knowledge_base = kbase)
        >>> w = Word([Letter("a")])
        >>> ot._ObservationTable__add_word_in_D(w)
        >>> ot._ObservationTable__add_word_in_D(w)
        Traceback (most recent call last):
        ...
        Exception: Word '[Letter('a')]' is already registered in D

        

        """
        if word is None:
            raise Exception("Word cannot be None")

        if word in self.D:
            raise Exception("Word '{}' is already registered in D".format(word))

        if word in self.ot_content.keys():
            raise Exception("Word '{}' is already registered in the content of the observation table".format(word))

        self._logger.debug("Registering word '{}' in D".format(word))
        self.D.append(word)

        # computes the value of all existing S and SA for the newly inserted word
        cels = dict()

        for word_in_S_or_SA in self.S + self.SA:
            # formulates a new OutputQuery and executes it
            output_query = OutputQuery(word_in_S_or_SA + word)
            self.__execute_query(output_query)
            if not output_query.is_queried():
                raise Exception("Query '{}' could not be queried".format(output_query))
            cels[word_in_S_or_SA] = output_query.output_word.last_letter()

        self.ot_content[word] = cels        
            
        
    def __add_word_in_S(self, word):
        """Add the specified word in S. In details, it does the following:
        - add the specified word in the list of words in S
        - executes the output queries (word + word_in_D) and store the results in DxS
        - recomputes new SAs made with the specified word + word_in_D        


        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> from pylstar.Letter import Letter
        >>> from pylstar.Word import Word
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable(input_letters = [], knowledge_base = kbase)
        >>> ot.initialize()
        >>> w = Word([Letter("a")])
        >>> ot._ObservationTable__add_word_in_S(w)

        """

        if word is None:
            raise Exception("Word cannot be None")

        if word in self.S:
            raise Exception("Word '{}' is already registered in S".format(word))

        if word in self.SA:
            raise Exception("Word '{}' is already registered in SA".format(word))

        self._logger.debug("Registering word '{}' in S".format(word))

        self.S.append(word)

        for word_in_D in self.D:

            cel = self.ot_content[word_in_D]
            # if word in cel.keys():
            #     raise Exception("Word '{}' already exists in observation table with D='{}'".format(word, word_in_D))
            # formulates a new OutputQuery and execute it            
            output_query = OutputQuery(word + word_in_D)
            self._logger.debug("Execute query : {}".format(output_query))
            self.__execute_query(output_query)
            if not output_query.is_queried():
                raise Exception("Query '{}' could not be queried".format(output_query))

            cel[word] = output_query.output_word.last_letter()

        for input_letter in self.input_letters:
            if isinstance(word.letters[0], EmptyLetter):
                new_word = Word([input_letter])
            else:
                new_word = word + Word([input_letter])
            self._logger.debug("Adding word: {}".format(new_word))            
            if new_word not in self.S:
                self.__add_word_in_SA(new_word)

    def __add_word_in_SA(self, word):
        """Add the specified word in SA. In details, it does the following:


        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> from pylstar.Letter import Letter
        >>> from pylstar.Word import Word
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable(input_letters = [], knowledge_base = kbase)
        >>> ot.initialize()
        >>> w = Word([Letter("a")])
        >>> ot._ObservationTable__add_word_in_SA(w)

        """

        if word is None:
            raise Exception("Word cannot be None")

        if word in self.SA:
            raise Exception("Word '{}' is already registered in SA".format(word))

        if word in self.S:
            raise Exception("Word '{}' is already registered in S".format(word))

        self._logger.debug("Registering word '{}' in SA".format(word))

        self.SA.append(word)

        for word_in_D in self.D:
            cel = self.ot_content[word_in_D]
            if word in cel.keys():
                raise Exception("Word '{}' already exists in observation table with D='{}'".format(word, word_in_D))
            # formulates a new OutputQuery and executes it
            output_query = OutputQuery(word + word_in_D)
            self._logger.debug("Execute query : {}".format(output_query))            
            self.__execute_query(output_query)
            if not output_query.is_queried():
                raise Exception("Query '{}' could not be queries".format(output_query))

            cel[word] = output_query.output_word.last_letter()
        
    def __execute_query(self, query):
        """This method triggers the execution of the specified query.
        
        An exception is raised if the query is None or if its execution failed

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> kbase = KnowledgeBase()
        >>> oTable = ObservationTable([], knowledge_base = kbase)
        >>> oTable._ObservationTable__execute_query(None)
        Traceback (most recent call last):
        ...
        Exception: Query cannot be None
        
        """
        if query is None:
            raise Exception("Query cannot be None")

        try:
            self.knowledge_base.resolve_query(query)
        except Exception as e:
            self._logger.error(e, exc_info=True)

    def build_hypothesis(self):
        """This method returns and Automata that follows the observation table.

        If the observation table is closed and consistent, it is possible to construct
        an hypothesis automata. Each state of the automata maps to a row of S in the observation table.        

        >>> from pylstar.ObservationTable import ObservationTable
        >>> from pylstar.KnowledgeBase import KnowledgeBase
        >>> from pylstar.Letter import Letter, EmptyLetter
        >>> from pylstar.Word import Word
        >>> l_lambda = EmptyLetter()
        >>> l_a = Letter('a')
        >>> l_b = Letter('b')
        >>> l_y = Letter('y')
        >>> l_z = Letter('z')
        >>> w_lambda = Word([l_lambda])
        >>> w_a = Word([l_a])
        >>> w_b = Word([l_b])
        >>> w_aa = Word([l_a, l_a])
        >>> w_ab = Word([l_a, l_b])
        >>> w_aaa = Word([l_a, l_a, l_a])
        >>> w_aab = Word([l_a, l_a, l_b])
        >>> kbase = KnowledgeBase()
        >>> ot = ObservationTable(input_letters = [l_a, l_b], knowledge_base = kbase)
        >>> ot.D.extend([w_a, w_b, w_aa])
        >>> ot.S.extend([w_lambda, w_a, w_aa])
        >>> ot.SA.extend([w_b, w_ab, w_aaa, w_aab])
        >>> ot.ot_content[w_a] = dict()
        >>> ot.ot_content[w_a][w_lambda] = l_z
        >>> ot.ot_content[w_a][w_a] = l_y
        >>> ot.ot_content[w_a][w_aa] = l_z
        >>> ot.ot_content[w_a][w_b] = l_y
        >>> ot.ot_content[w_a][w_ab] = l_z
        >>> ot.ot_content[w_a][w_aaa] = l_z
        >>> ot.ot_content[w_a][w_aab] = l_y
        >>> ot.ot_content[w_b] = dict()
        >>> ot.ot_content[w_b][w_lambda] = l_z
        >>> ot.ot_content[w_b][w_a] = l_y
        >>> ot.ot_content[w_b][w_aa] = l_z
        >>> ot.ot_content[w_b][w_b] = l_y
        >>> ot.ot_content[w_b][w_ab] = l_z
        >>> ot.ot_content[w_b][w_aaa] = l_z
        >>> ot.ot_content[w_b][w_aab] = l_y
        >>> ot.ot_content[w_aa] = dict()
        >>> ot.ot_content[w_aa][w_lambda] = l_y
        >>> ot.ot_content[w_aa][w_a] = l_z
        >>> ot.ot_content[w_aa][w_aa] = l_z
        >>> ot.ot_content[w_aa][w_b] = l_z
        >>> ot.ot_content[w_aa][w_ab] = l_y
        >>> ot.ot_content[w_aa][w_aaa] = l_y
        >>> ot.ot_content[w_aa][w_aab] = l_z        
        >>> print(ot)
                                                | [Letter('a')] | [Letter('b')] | [Letter('a'), Letter('a')]
        --------------------------------------- | ------------- | ------------- | --------------------------
        [EmptyLetter]                           | Letter('z')   | Letter('z')   | Letter('y')               
        [Letter('a')]                           | Letter('y')   | Letter('y')   | Letter('z')               
        [Letter('a'), Letter('a')]              | Letter('z')   | Letter('z')   | Letter('z')               
        ~~~                                     | ~~~           | ~~~           | ~~~                       
        [Letter('b')]                           | Letter('y')   | Letter('y')   | Letter('z')               
        [Letter('a'), Letter('b')]              | Letter('z')   | Letter('z')   | Letter('y')               
        [Letter('a'), Letter('a'), Letter('a')] | Letter('z')   | Letter('z')   | Letter('y')               
        [Letter('a'), Letter('a'), Letter('b')] | Letter('y')   | Letter('y')   | Letter('z')               
        --------------------------------------- | ------------- | ------------- | --------------------------
        >>> print(ot.is_closed())
        True
        >>> print(ot.find_inconsistency())
        None
        >>> automata = ot.build_hypothesis()
        >>> print(automata.build_dot_code())
        digraph "Automata" {
        "0" [shape=doubleoctagon, style=filled, fillcolor=white, URL="0"];
        "1" [shape=ellipse, style=filled, fillcolor=white, URL="1"];
        "2" [shape=ellipse, style=filled, fillcolor=white, URL="2"];
        "0" -> "1" [fontsize=5, label="a / z", URL="t0"];
        "0" -> "1" [fontsize=5, label="b / z", URL="t1"];
        "1" -> "2" [fontsize=5, label="a / y", URL="t2"];
        "1" -> "0" [fontsize=5, label="b / y", URL="t3"];
        "2" -> "0" [fontsize=5, label="a / z", URL="t4"];
        "2" -> "1" [fontsize=5, label="b / z", URL="t5"];
        }
        
        """

        states = []
        transitions = []
        initial_state = None
        words_and_states = []
        long_state_name_to_states = dict()        
        
        # find all rows in S
        rows_in_S = [(s, self.__get_row(s)) for s in self.S]

        # get all unique rows
        S_with_same_rows = collections.OrderedDict()
        for word_in_s, row_s in rows_in_S:
            key = ','.join([str(w) for w in row_s])
            if key not in S_with_same_rows.keys():
                S_with_same_rows[key] = list()
            S_with_same_rows[key].append(word_in_s)        

        # build the list of states of the hypothesis (and identify the initial state)
        for long_state_name, words_in_S in S_with_same_rows.items():
            state_name = str(len(states)) #''.join(long_state_name.replace("Letter(", "").replace(')', ''))
            state = State(name = state_name)
            states.append(state)

            words_and_states.append((words_in_S[0], state))
            long_state_name_to_states[long_state_name] = state
            
            # check if its the initial state
            epsilon_word_found = Word([EmptyLetter()]) in words_in_S
            self._logger.debug("state  :{} , {} / {}".format(long_state_name, words_in_S, epsilon_word_found))
            
            if initial_state is None and epsilon_word_found:
                initial_state = state
            elif epsilon_word_found and initial_state is not None:
                raise Exception("Multiple initial state found.")

        if initial_state is None:
            raise Exception("Can't find any initial state")

        # computes the transitions for each state of the automata
        for word, state in words_and_states:

            for input_letter in self.input_letters:

                # computes the output state
                new_word = word + Word([input_letter])
                row_new_word = self.__get_row(new_word)

                output_state_name = ','.join([str(w) for w in row_new_word])
                if output_state_name not in long_state_name_to_states.keys():
                    for x in long_state_name_to_states.keys():
                        self._logger.debug(x)
                
                    raise Exception("Cannot find a state with following name : '{}'".format(output_state_name))

                output_state = long_state_name_to_states[output_state_name]
                output_letter = self.ot_content[Word([input_letter])][word]

                transition_name = "t{}".format(len(transitions))
                transition = Transition(name = transition_name,
                                        output_state = output_state,
                                        input_letter = input_letter,
                                        output_letter = output_letter)
                state.transitions.append(transition)
                transitions.append(transition)
                
                                                    
        return Automata(initial_state = initial_state)

    def __str__(self):
        result = []

        # lets build a matrix out of the observation table
        # a matrix : a list of list a.k.a a list of rows
        matrix = []

        # adding header
        header_row = [''] + [str(d) for d in self.D]
        matrix.append(header_row)
        
        for row_name in self.S:
            row_result = [str(row_name)] + [str(w) for w in self.__get_row(row_name)]
            matrix.append(row_result)

        matrix.append(['~~~']*len(header_row))

        for row_name in self.SA:
            row_result = [str(row_name)] + [str(w) for w in self.__get_row(row_name)]
            matrix.append(row_result)

        self._logger.debug(matrix)

        cs = zip(*matrix)
        c_ws = [max(len(value) for value in c) for c in cs]
        line = ["-"*w for w in c_ws]
        matrix.insert(1, line)
        matrix.append(line)
        format = ' | '.join(['%%-%ds' % w for w in c_ws])
        # Format data
        result = [(format % tuple(r)) for r in matrix]
        return '\n'.join(result)        

    @property
    def knowledge_base(self):
        """The knowledge base plays the role of a proxy between the algorithm and the System Under Learning.
        """
        return self.__knowledge_base
    
    @knowledge_base.setter
    def knowledge_base(self, knowledge_base):
        if knowledge_base is None:
            raise Exception("Knowledge base cannot be None")
        
        self.__knowledge_base = knowledge_base

        
        
        
            

        

        



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import abc

# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.KnowledgeTree import KnowledgeTree
from pylstar.KnowledgeBaseStats import KnowledgeBaseStats


@PylstarLogger
class KnowledgeBase(object):
    """An abstract class that implements a knowledge base.

    This knowledge base stores all the query results in a tree.

    >>> from pylstar.KnowledgeBase import KnowledgeBase
    >>> from pylstar.OutputQuery import OutputQuery
    >>> from pylstar.Word import Word
    >>> from pylstar.Letter import Letter
    >>> word1 = Word([Letter('a'), Letter('b')])
    >>> query1 = OutputQuery(word1)
    >>> kbase = KnowledgeBase()
    >>> kbase.resolve_query(query1)
    Traceback (most recent call last):
    ...
    Exception: Passive inference process
    >>> word2 = Word([Letter('a'), Letter('b'), Letter('c')])
    >>> word3 = Word([Letter('1'), Letter('2'), Letter('3')])
    >>> kbase.add_word(input_word = word2, output_word = word3)
    >>> word4 = Word([Letter('a'), Letter('d'), Letter('e')])
    >>> word5 = Word([Letter('1'), Letter('4'), Letter('5')])
    >>> kbase.add_word(input_word = word4, output_word = word5)
    >>> word6 = Word([Letter('f'), Letter('g')])
    >>> word7 = Word([Letter('6'), Letter('7')])
    >>> kbase.add_word(input_word = word6, output_word = word7)
    >>> kbase.resolve_query(query1)
    >>> print(query1.output_word)
    [Letter('1'), Letter('2')]
    >>> word8 = Word([Letter('a'), Letter('d'), Letter('e')])
    >>> query2 = OutputQuery(word8)
    >>> kbase.resolve_query(query2)
    >>> print(query2.output_word)
    [Letter('1'), Letter('4'), Letter('5')]
    



    
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, cache_file_path = None):
        self.knowledge_tree = KnowledgeTree(cache_file_path = cache_file_path)
        self.stats = KnowledgeBaseStats()

    def load_cache(self, possible_letters):
        self.knowledge_tree.load_cache(possible_letters)

    def write_cache(self):
        self.knowledge_tree.write_cache()

    def __str__(self):
        return str(self.knowledge_tree)

    def resolve_query(self, query):
        """This method can be use to interogate the cache for
        the output word associated with the specified query. If no
        previous knowledge can be found for this query, the input word
        is submitted to the target.

        """
        if query is None:
            raise Exception("Query cannot be None")

        query.output_word = self._resolve_word(query.input_word)

    def _resolve_word(self, word):
        if word is None:
            raise Exception("Word cannot be None")

        try:
            self.stats.nb_query += 1
            self.stats.nb_letter += len(word.letters)

            return self.knowledge_tree.get_output_word(word)
        except Exception:        
            self._logger.debug("Knowledge base has no previous knowledge for '{}'".format(word))

            self.stats.nb_submited_query += 1
            self.stats.nb_submited_letter += len(word.letters)

            output = self._execute_word(word)
            
            if output is not None:
                self.knowledge_tree.add_word(input_word = word, output_word = output)
            return output
    

    def _execute_word(self, word):
        """This method must be overwritten by subclasses that implements
        an active learning process.
        """
        raise Exception("Passive inference process")
        
    def add_word(self, input_word, output_word):
        """This method stores in the knowledge base the relationship between
        the specified input_word and output_word
        """
        self._logger.debug("adding : {}".format(','.join([str(l) for l in input_word.letters])))
        self.knowledge_tree.add_word(input_word, output_word)

        

        
        

        
        

    



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import abc

# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.KnowledgeBase import KnowledgeBase


@PylstarLogger
class ActiveKnowledgeBase(KnowledgeBase):
    """An abstract class that implements the main mecanism of an active knwoledge base.

    

    """

    def __init__(self, cache_file_path = None):
        super(ActiveKnowledgeBase, self).__init__(cache_file_path = cache_file_path)

    def _execute_word(self, word):
        """Executes the specified word."""
        
        if word is None:
            raise Exception("Word cannot be None")
        
        self._logger.debug("Execute word '{}'".format(word))

        self.start_target()
        try:
            return self.submit_word(word)
        finally:
            self.stop_target()

    @abc.abstractmethod
    def start_target(self):
        raise NotImplementedError()

    @abc.abstractmethod    
    def stop_target(self):
        raise NotImplementedError()

    @abc.abstractmethod    
    def submit_word(self, word):
        raise NotImplementedError()

        
        

        

    



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------


# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.Letter import EmptyLetter

@PylstarLogger
class Word(object):
    """A word represents a sequence of letters.

    >>> from pylstar.Word import Word
    >>> from pylstar.Letter import Letter
    >>> l1 = Letter("a")
    >>> l2 = Letter("b")
    >>> w1 = Word([l1, l2])
    >>> print(w1)
    [Letter('a'), Letter('b')]

    
    """

    def __init__(self, letters = None, normalize=True):

        if normalize:
            self.letters = letters
        else:
            self.letters = []
            for l in letters:
                self.letters.append(l)

    def __hash__(self):
        return hash(repr(self))
    
    def __eq__(self, other):
        if not isinstance(other, Word):
            return False
        return self.letters == other.letters

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "[{}]".format(", ".join([str(letter) for letter in self.letters]))

    def __repr__(self):
        return "[{}]".format(", ".join([str(letter) for letter in self.letters]))

    def __len__(self):
        return len(self.letters)

    def last_letter(self):
        return self.letters[-1]

    def __add__(self, other):
        """Two words can be appended to produce a new one

        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> l1 = Letter("a")
        >>> l2 = Letter("b")
        >>> l3 = Letter("c")
        >>> l4 = Letter("d")
        >>> w1 = Word([l1, l2])
        >>> w2 = Word([l3, l4])
        >>> w3 = w1 + w2
        >>> print(w3)
        [Letter('a'), Letter('b'), Letter('c'), Letter('d')]


        Only two words can be added

        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> l1 = Letter("a")
        >>> w1 = Word([l1])
        >>> w2 = "data"
        >>> w3 = w1 + w2
        Traceback (most recent call last):
        ...
        Exception: Only two words can be added
        
        """

        if not isinstance(other, Word):
            raise Exception("Only two words can be added")

        if len(self.letters) >= 1 and isinstance(self.letters[0], EmptyLetter):
            return Word(self.letters[1:] + other.letters)
            
        return Word(self.letters + other.letters)

    @property
    def letters(self):
        """Letters that are embeded in the current word"""
        return self.__letters
    
    @letters.setter
    def letters(self, letters):
        if letters is None:
            letters = []

        if len(letters) > 1 and isinstance(letters[0], EmptyLetter):
            letters = letters[1:]
        
        self.__letters = []        
        for letter in letters:            
            self.__letters.append(letter)



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------


# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger


@PylstarLogger
class State(object):
    """Definition of a state that belongs to an automata

    >>> from pylstar.automata.State import State
    >>> from pylstar.automata.Transition import Transition
    >>> s0 = State("s0")
    >>> s1 = State("s1")
    >>> print(s0)
    s0
    >>> print(s0.transitions)
    []

    """

    def __init__(self, name):
        self.name = name
        self.transitions = []

    def __str__(self):
        return self.name
        
    def visit(self, input_letter):
        """This method computes which transition can be triggered given the
        specified input_letter. It returns a tupple made of the output letter
        that is attached to the found transition and the state it reaches.

        >>> from pylstar.automata.State import State
        >>> from pylstar.automata.Transition import Transition
        >>> from pylstar.Letter import Letter
        >>> la = Letter('a')
        >>> l0 = Letter('0')
        >>> lb = Letter('b')
        >>> l1 = Letter('1')
        >>> s0 = State("s0")
        >>> s1 = State("s1")
        >>> t0 = Transition("t0", s1, la, l0)
        >>> t1 = Transition("t1", s0, lb, l1)
        >>> s0.transitions = [t0, t1]
        >>> (output_letter, output_state) = s0.visit(la)
        >>> print(output_letter)
        Letter('0')
        >>> print(output_state)
        s1
        >>> (output_letter, output_state) = s0.visit(lb)
        >>> print(output_letter)
        Letter('1')
        >>> print(output_state)
        s0
        >>> s0.visit(None)
        Traceback (most recent call last):
        ...
        Exception: input letter cannot be None
        >>> s0.visit(l0)
        Traceback (most recent call last):
        ...
        Exception: No transition in state 's0' could be found given letter 'Letter('0')' 


        """
        
        if input_letter is None:
            raise Exception("input letter cannot be None")

        for transition in self.transitions:
            if transition.input_letter == input_letter:
                return (transition.output_letter, transition.output_state)

        raise Exception("No transition in state '{}' could be found given letter '{}' ".format(self.name, input_letter))
                
        

    



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import random

# +----------------------------------------------------------------------------
# | pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.Word import Word
from pylstar.OutputQuery import OutputQuery


@PylstarLogger
class RandomWalkMethod(object):
    """This algorithm performs a random walk accross the hypothesis states.
    The output symbols produced while triggering transitions are compared to the ones produced when trigerring
    the same transitions in the targeted automata.

    >>> from pylstar.LSTAR import LSTAR
    >>> from pylstar.automata.State import State
    >>> from pylstar.automata.Transition import Transition
    >>> from pylstar.automata.Automata import Automata
    >>> from pylstar.Letter import Letter
    >>> from pylstar.FakeActiveKnowledgeBase import FakeActiveKnowledgeBase
    >>> from pylstar.eqtests.RandomWalkMethod import RandomWalkMethod
    >>> symbol_a = "a"
    >>> symbol_b = "b"
    >>> symbol_c = "c"
    >>> symbol_1 = 1
    >>> symbol_2 = 2
    >>> symbol_3 = 3
    >>> l_a = Letter(symbol_a)
    >>> l_b = Letter(symbol_b)
    >>> l_c = Letter(symbol_c)
    >>> l_1 = Letter(symbol_1)
    >>> l_2 = Letter(symbol_2)
    >>> l_3 = Letter(symbol_3)
    >>> s0 = State("S0")
    >>> s1 = State("S1")
    >>> s2 = State("S2")
    >>> t1 = Transition("t1", output_state=s0, input_letter=l_a, output_letter=l_1)
    >>> t2 = Transition("t2", output_state=s1, input_letter=l_b, output_letter=l_2)
    >>> t3 = Transition("t3", output_state=s2, input_letter=l_c, output_letter=l_3)
    >>> s0.transitions = [t1, t2, t3]
    >>> t4 = Transition("t4", output_state=s1, input_letter=l_a, output_letter=l_2)
    >>> t5 = Transition("t5", output_state=s1, input_letter=l_b, output_letter=l_3)
    >>> t6 = Transition("t6", output_state=s0, input_letter=l_c, output_letter=l_1)
    >>> s1.transitions = [t4, t5, t6]
    >>> t7 = Transition("t7", output_state=s2, input_letter=l_a, output_letter=l_2)
    >>> t8 = Transition("t8", output_state=s2, input_letter=l_b, output_letter=l_3)
    >>> t9 = Transition("t9", output_state=s1, input_letter=l_c, output_letter=l_1)
    >>> s2.transitions = [t7, t8, t9]
    >>> automata = Automata(s0)
    >>> kbase = FakeActiveKnowledgeBase(automata)
    >>> input_vocabulary = [symbol_a, symbol_b, symbol_c]
    >>> input_letters = [Letter(s) for s in input_vocabulary]
    >>> eqTests = RandomWalkMethod(kbase, input_letters, 10000, 0.75)
    >>> lstar = LSTAR(input_vocabulary, kbase, max_states = 5, eqtests=eqTests)
    >>> infered_automata = lstar.learn()
    >>> print(infered_automata.build_dot_code())
    digraph "Automata" {
    "0" [shape=doubleoctagon, style=filled, fillcolor=white, URL="0"];
    "2" [shape=ellipse, style=filled, fillcolor=white, URL="2"];
    "1" [shape=ellipse, style=filled, fillcolor=white, URL="1"];
    "0" -> "0" [fontsize=5, label="a / 1", URL="t0"];
    "0" -> "1" [fontsize=5, label="b / 2", URL="t1"];
    "0" -> "2" [fontsize=5, label="c / 3", URL="t2"];
    "2" -> "2" [fontsize=5, label="a / 2", URL="t6"];
    "2" -> "2" [fontsize=5, label="b / 3", URL="t7"];
    "2" -> "1" [fontsize=5, label="c / 1", URL="t8"];
    "1" -> "1" [fontsize=5, label="a / 2", URL="t3"];
    "1" -> "1" [fontsize=5, label="b / 3", URL="t4"];
    "1" -> "0" [fontsize=5, label="c / 1", URL="t5"];
    }

    
    >>> from pylstar.LSTAR import LSTAR
    >>> from pylstar.automata.State import State
    >>> from pylstar.automata.Transition import Transition
    >>> from pylstar.automata.Automata import Automata
    >>> from pylstar.Letter import Letter
    >>> from pylstar.FakeActiveKnowledgeBase import FakeActiveKnowledgeBase
    >>> # input symbols
    >>> symbol_hello = "hello"
    >>> symbol_bye = "bye"
    >>> symbol_pass_valid = "pass valid"
    >>> symbol_pass_invalid = "pass invalid"
    >>> symbol_cmd1 = "cmd1"
    >>> symbol_cmd2 = "cmd2"
    >>> # output symbols
    >>> symbol_pass_request = "pass?"
    >>> symbol_ack = "ack"
    >>> symbol_welcome = "welcome"
    >>> symbol_error = "error"
    >>> # create a letter for each symbol
    >>> l_hello = Letter(symbol_hello)
    >>> l_bye = Letter(symbol_bye)
    >>> l_pass_valid = Letter(symbol_pass_valid)
    >>> l_pass_invalid = Letter("pass invalid")
    >>> l_cmd1 = Letter(symbol_cmd1)
    >>> l_cmd2 = Letter(symbol_cmd2)
    >>> l_welcome = Letter(symbol_welcome)
    >>> l_ack = Letter(symbol_ack)
    >>> l_pass_request = Letter(symbol_pass_request)
    >>> l_error = Letter(symbol_error)
    >>> # create the infered automata
    >>> s0 = State("S0")
    >>> s1 = State("S1")
    >>> s2 = State("S2")
    >>> t1 = Transition("t1", output_state=s1, input_letter=l_hello, output_letter=l_pass_request)
    >>> t2 = Transition("t2", output_state=s0, input_letter=l_bye, output_letter=l_ack)
    >>> t3 = Transition("t3", output_state=s0, input_letter=l_pass_valid, output_letter=l_error)
    >>> t4 = Transition("t4", output_state=s0, input_letter=l_pass_invalid, output_letter=l_error)
    >>> t5 = Transition("t5", output_state=s0, input_letter=l_cmd1, output_letter=l_error)
    >>> t6 = Transition("t6", output_state=s0, input_letter=l_cmd2, output_letter=l_error)
    >>> s0.transitions = [t1, t2, t3, t4, t5, t6]
    >>> t7 = Transition("t7", output_state=s1, input_letter=l_hello, output_letter=l_error)
    >>> t8 = Transition("t8", output_state=s0, input_letter=l_bye, output_letter=l_ack)
    >>> t9 = Transition("t9", output_state=s2, input_letter=l_pass_valid, output_letter=l_welcome)
    >>> t10 = Transition("t10", output_state=s1, input_letter=l_pass_invalid, output_letter=l_error)
    >>> t11 = Transition("t11", output_state=s1, input_letter=l_cmd1, output_letter=l_error)
    >>> t12 = Transition("t12", output_state=s1, input_letter=l_cmd2, output_letter=l_error)
    >>> s1.transitions = [t7, t8, t9, t10, t11, t12]
    >>> t13 = Transition("t13", output_state=s2, input_letter=l_hello, output_letter=l_error)
    >>> t14 = Transition("t14", output_state=s0, input_letter=l_bye, output_letter=l_ack)
    >>> t15 = Transition("t15", output_state=s2, input_letter=l_pass_valid, output_letter=l_error)
    >>> t16 = Transition("t16", output_state=s2, input_letter=l_pass_invalid, output_letter=l_error)
    >>> t17 = Transition("t17", output_state=s2, input_letter=l_cmd1, output_letter=l_ack)
    >>> t18 = Transition("t18", output_state=s2, input_letter=l_cmd2, output_letter=l_ack)
    >>> s2.transitions = [t13, t14, t15, t16, t17, t18]
    >>> automata = Automata(s0)
    >>> kbase = FakeActiveKnowledgeBase(automata)
    >>> input_vocabulary = [symbol_hello, symbol_bye, symbol_pass_valid, symbol_pass_invalid, symbol_cmd1, symbol_cmd2]
    >>> input_letters = [Letter(s) for s in input_vocabulary]
    >>> eqTests = RandomWalkMethod(kbase, input_letters, 10000, 0.75)
    >>> lstar = LSTAR(input_vocabulary, kbase, max_states = 5, eqtests=eqTests)
    >>> infered_automata = lstar.learn()
    >>> print(infered_automata.build_dot_code())
    digraph "Automata" {
    "0" [shape=doubleoctagon, style=filled, fillcolor=white, URL="0"];
    "1" [shape=ellipse, style=filled, fillcolor=white, URL="1"];
    "2" [shape=ellipse, style=filled, fillcolor=white, URL="2"];
    "0" -> "1" [fontsize=5, label="hello / pass?", URL="t0"];
    "0" -> "0" [fontsize=5, label="bye / ack", URL="t1"];
    "0" -> "0" [fontsize=5, label="pass valid / error", URL="t2"];
    "0" -> "0" [fontsize=5, label="pass invalid / error", URL="t3"];
    "0" -> "0" [fontsize=5, label="cmd1 / error", URL="t4"];
    "0" -> "0" [fontsize=5, label="cmd2 / error", URL="t5"];
    "1" -> "1" [fontsize=5, label="hello / error", URL="t6"];
    "1" -> "0" [fontsize=5, label="bye / ack", URL="t7"];
    "1" -> "2" [fontsize=5, label="pass valid / welcome", URL="t8"];
    "1" -> "1" [fontsize=5, label="pass invalid / error", URL="t9"];
    "1" -> "1" [fontsize=5, label="cmd1 / error", URL="t10"];
    "1" -> "1" [fontsize=5, label="cmd2 / error", URL="t11"];
    "2" -> "2" [fontsize=5, label="hello / error", URL="t12"];
    "2" -> "0" [fontsize=5, label="bye / ack", URL="t13"];
    "2" -> "2" [fontsize=5, label="pass valid / error", URL="t14"];
    "2" -> "2" [fontsize=5, label="pass invalid / error", URL="t15"];
    "2" -> "2" [fontsize=5, label="cmd1 / ack", URL="t16"];
    "2" -> "2" [fontsize=5, label="cmd2 / ack", URL="t17"];
    }



    """
    

    def __init__(self, knowledge_base, input_letters, max_steps, restart_probability):
        self.knowledge_base = knowledge_base
        self.input_letters = input_letters
        self.max_steps = max_steps
        self.restart_probability = restart_probability

    def find_counterexample(self, hypothesis):
        if hypothesis is None:
            raise Exception("Hypothesis cannot be None")

        
        self._logger.info("Starting the RandomWalk Algorithm to search for a counter-example")
        
        i_step = 0
        first_step_after_restart = True
        current_state = hypothesis.initial_state

        input_word = Word()
        hypothesis_output_word = Word()
        force_restart = False
        while i_step < self.max_steps:

            # should we restart
            if not first_step_after_restart:
                if force_restart or random.random() < self.restart_probability:
                    current_state = hypothesis.initial_state
                    first_step_after_restart = True

                    counterexample_query = self.__check_equivalence(input_word, hypothesis_output_word)
                    if counterexample_query is not None:
                        return counterexample_query
                    
                    input_word = Word()
                    hypothesis_output_word = Word()
                    force_restart = False
            else:
                first_step_after_restart = False

            try:
                (new_state, input_letter, output_letter) = self.__walk(current_state)
                current_state = new_state
                input_word.letters.append(input_letter)
                hypothesis_output_word.letters.append(output_letter)
                
            except Exception as e:
                self._logger.warn(e)
                force_restart = True

            i_step += 1

    def __check_equivalence(self, input_word, expected_output_word):
        if input_word is None:
            raise Exception("Input word cannot be None")
            
        if expected_output_word is None:
            raise Exception("Expected word cannot be None")

        query = OutputQuery(input_word)
        self.knowledge_base.resolve_query(query)

        if query.output_word != expected_output_word:
            self._logger.info("Found a counter-example : input: '{}', expected: '{}', observed: '{}'".format(input_word, expected_output_word, query.output_word))
            return query
        return None

    def __walk(self, current_state):
        if current_state is None:
            raise Exception("Current state cannot be None")

        picked_transition = random.choice(current_state.transitions)
        if picked_transition is None:
            raise Exception("Found a state that accepts no more transition")
        
        (output_letter, output_state) = current_state.visit(picked_transition.input_letter)

        return (output_state, picked_transition.input_letter, output_letter)

        



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------


# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger


@PylstarLogger
class OutputQuery(object):
    """An output query is a query made of an input word and an output word.
    An input word is the word that is sent to the teacher which answers with the output word

    Input and output words always have the same size, EmptyLetters can be inserted to ensure this.

    An output query cannot be created without specifing its input word
    
    >>> from pylstar.OutputQuery import OutputQuery
    >>> output_query = OutputQuery(None)
    Traceback (most recent call last):
    ...
    Exception: Input word cannot be None


    """
    

    def __init__(self, word):
        self.input_word = word
        self.output_word = None

    def is_queried(self):
        """This method returns True is the query was queried

        >>> from pylstar.OutputQuery import OutputQuery
        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> w1 = Word([Letter("a")])
        >>> output_query = OutputQuery(w1)
        >>> output_query.is_queried()
        False
        >>> print(output_query)
        OutputQuery(I = [Letter('a')], O = None)

        """
        return self.output_word is not None

    def __str__(self):
        return "OutputQuery(I = {}, O = {})".format(self.input_word, self.output_word)

    def multiply(self, queries):
        if queries is None:
            raise Exception("Queries cannot be None")

        return [OutputQuery(self.input_word + query.input_word) for query in queries]
        
    @property
    def input_word(self):
        """Input word that makes the query"""
        return self.__input_word
    
    @input_word.setter
    def input_word(self, input_word):
        if input_word is None:
            raise Exception("Input word cannot be None")
        
        self.__input_word = input_word


        
        



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
from distutils.core import Command
import os
import sys
import unittest


class test_command(Command):
    description = "Test PYLSTAR"

    user_options = [('reportfile=', None, 'name of the generated XML report file (not required)') ]

    def initialize_options(self):
        self.reportfile = None
        self._dir = os.getcwd()

    def finalize_options(self):
        pass

    def run(self):
        '''
        Finds all the tests modules in test/, and runs them.
        '''
        sys.path.insert(0, 'src/')

        sys.path.insert(0, 'test/src/')

        from common.xmlrunner import XMLTestRunner
        from test_pylstar import suite_global

        # We retrieve the current test suite
        currentTestSuite = suite_global.getSuite()

        if self.reportfile is None or len(self.reportfile) == 0:
            runner = unittest.TextTestRunner()
            runner.run(currentTestSuite)
        else:
            # We execute the test suite
            File = open(self.reportfile, 'w')
            File.write('<?xml version="1.0" encoding="utf-8"?>\n')
            reporter = XMLTestRunner(File)
            reporter.run(currentTestSuite)
            File.close()

            self.cleanFile(self.reportfile)

    def cleanFile(self, filePath):
        """Clean the file to handle non-UTF8 bytes.
        """

        aFile = open(filePath, 'r')
        data = aFile.read()
        aFile.close()

        cleanData = ""
        for c in data:
            if (0x1f < ord(c) < 0x80) or (ord(c) == 0x9) or (ord(c) == 0xa) or (ord(c) == 0xd):
                cleanData += c
            else:
                cleanData += repr(c)

        aFile = open(filePath, 'w')
        aFile.write(cleanData)
        aFile.close()



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import os
import json

# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.Word import Word
from pylstar.Letter import Letter


@PylstarLogger
class KnowledgeNode(object):

    def __init__(self, input_letter, output_letter):
        self.input_letter = input_letter
        self.output_letter = output_letter
        self.children = dict()

    def __str__(self, level=0):
        return json.dumps(self.serialize(), sort_keys=True, indent=4, separators=(',', ': '))

    def serialize(self):
        """This method return a serialized representation of the node"""
        node = {
            "input_letter" : self.input_letter.serialize(),
            "output_letter": self.output_letter.serialize(),
            "children" : [c.serialize() for c in self.children.values()]
        }
        return node
    
    @staticmethod
    def deserialize(dict_data, possible_letters):
        if dict_data is None:
            raise Exception("dict_data cannot be None")
        input_letter = Letter.deserialize(dict_data['input_letter'], possible_letters)
        output_letter = Letter.deserialize(dict_data['output_letter'], possible_letters)
        node = KnowledgeNode(input_letter, output_letter)
        for child in dict_data["children"]:
            child_node = KnowledgeNode.deserialize(child, possible_letters)
            node.children[child_node.input_letter] = child_node

        return node

    def traverse(self, input_letters, output_letters = None):

        if input_letters[0] != self.input_letter:
            raise Exception("Node cannot be traversed with input letter '{}'".format(input_letters[0]))
        if output_letters is not None and output_letters[0] != self.output_letter:
            raise Exception("Node '{}' cannot be traversed with output letter '{}'".format(self, output_letters[0]))
        if output_letters is not None and len(input_letters) != len(output_letters):
            raise Exception("Specified input and output letters do not have the same length")


        if len(input_letters) < 2:
            return [self.output_letter]

        current_input_letter = input_letters[1]
        current_output_letter = None
        if output_letters is not None:
            current_output_letter = output_letters[1]


        if current_input_letter in self.children:
            child = self.children[current_input_letter]
            
            if current_output_letter is not None and child.output_letter != current_output_letter:
                raise Exception("Incompatible path found, expected '{}' found '{}'".format(child.output_letter.symbols, current_output_letter.symbols))

            if output_letters is None:
                new_output_letters = None
            else:
                new_output_letters = output_letters[1:]
                
            new_input_letters = input_letters[1:]            

            return [self.output_letter] + child.traverse(new_input_letters, output_letters = new_output_letters)
        
        elif output_letters is not None:
            new_children = KnowledgeNode(input_letter = input_letters[1], output_letter = output_letters[1])
            self.children[new_children.input_letter] = new_children
            new_input_letters = input_letters[1:]
            new_output_letters = output_letters[1:]
            return [self.output_letter] + new_children.traverse(new_input_letters, output_letters = new_output_letters)

        raise Exception("Cannot traverse node '{}' with subsequences '{}'".format(self, ', '.join([str(l) for l in input_letters])))

    @property
    def input_letter(self):
        """Input letter"""
        return self.__input_letter

    @input_letter.setter
    def input_letter(self, input_letter):
        if input_letter is None:
            raise Exception("Input letter cannot be None")
        self.__input_letter = input_letter

    @property
    def output_letter(self):
        """Output letter"""
        return self.__output_letter

    @output_letter.setter
    def output_letter(self, output_letter):
        if output_letter is None:
            raise Exception("Output letter cannot be None")
        self.__output_letter = output_letter


@PylstarLogger
class KnowledgeTree(object):
    """A pythonic implementation of a tree that hosts query results.

    >>> from pylstar.KnowledgeTree import KnowledgeTree
    >>> from pylstar.Word import Word
    >>> from pylstar.Letter import Letter
    >>> tree = KnowledgeTree()
    >>> input_word = Word([Letter("a"), Letter("b")])
    >>> output_word = Word([Letter(1), Letter(2)])
    >>> tree.get_output_word(input_word)
    Traceback (most recent call last):
    ...
    Exception: No path found
    >>> tree.add_word(input_word, output_word)
    >>> print(tree.get_output_word(input_word))
    [Letter(1), Letter(2)]
    >>> eq_input_word = Word([Letter("a"), Letter("b")])
    >>> print(tree.get_output_word(eq_input_word))
    [Letter(1), Letter(2)]
    """

    def __init__(self, cache_file_path = None):
        self.__cache_file_path = cache_file_path
        self.__nb_added_word = 0
        self.roots = []

    def __str__(self):
        result = '\n'.join([root.__str__(level=1).rstrip() for root in self.roots])
        return 'Tree (\n{}\n)'.format(result)


    def get_output_word(self, input_word):
        if input_word is None:
            raise Exception("Input word cannot be None")

        for root in self.roots:
            try:
                w = Word(root.traverse(input_word.letters))
                self._logger.info("I = {} > O = {}".format(input_word, w))
                return w
            except Exception:
                pass

        raise Exception("No path found")

    def add_word(self, input_word, output_word):
        """This method can be use to associate an input word to an output word

        >>> from pylstar.KnowledgeTree import KnowledgeTree
        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> tree = KnowledgeTree()
        >>> input_word = Word([Letter("a"), Letter("b")])
        >>> output_word = Word([Letter(1), Letter(2)])
        >>> tree.add_word(input_word, output_word)

        The same association can be inserted twice iif both input and output are equivalent
        to the previously inserted association

        >>> from pylstar.KnowledgeTree import KnowledgeTree
        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> tree = KnowledgeTree()
        >>> input_word = Word([Letter("a"), Letter("b")])
        >>> output_word = Word([Letter(1), Letter(2)])
        >>> tree.add_word(input_word, output_word)
        >>> input_word2 = Word([Letter("a"), Letter("b")])
        >>> output_word2 = Word([Letter(1), Letter(2)])
        >>> tree.add_word(input_word2, output_word2)
        >>> output_word3 = Word([Letter(1), Letter(1)])
        >>> tree.add_word(input_word2, output_word3)
        Traceback (most recent call last):
        ...
        Exception: Incompatible path found, expected '{2}' found '{1}'

        """

        if input_word is None:
            raise Exception("Input word cannot be None")
        if output_word is None:
            raise Exception("Output word cannot be None")
        if len(input_word) != len(output_word):
            raise Exception("Input and output words do not have the same size")

        self.__add_letters(input_word.letters, output_word.letters)

        self.__nb_added_word += 1
        if self.__cache_file_path is not None and self.__nb_added_word % 100 == 0:
            self.write_cache()

    def write_cache(self):
        """This method writes the content of the knowledge tree to the self.cache_file_path.

        >>> cache_file = "/tmp/test_ktree_cache.dump"
        >>> from pylstar.KnowledgeTree import KnowledgeTree
        >>> from pylstar.Word import Word
        >>> from pylstar.Letter import Letter
        >>> l_a = Letter("a")
        >>> l_b = Letter("b")
        >>> l_c = Letter("c")
        >>> l_1 = Letter(1)
        >>> l_2 = Letter(2)
        >>> l_3 = Letter(3)
        >>> tree = KnowledgeTree(cache_file_path = cache_file)
        >>> input_word = Word([l_a, l_b])
        >>> output_word = Word([l_1, l_2])
        >>> tree.add_word(input_word, output_word)
        >>> tree.write_cache()        
        >>> input_word = Word([l_a, l_c])
        >>> output_word = Word([l_1, l_3])
        >>> tree.add_word(input_word, output_word)
        >>> tree.write_cache()
        >>> tree2 = KnowledgeTree(cache_file_path = cache_file)
        >>> tree2.load_cache(possible_letters = [l_a, l_b, l_c, l_1, l_2, l_3])
        >>> print(tree2.get_output_word(input_word))
        [Letter(1), Letter(3)]
        
        """

        if self.__cache_file_path is None:
            raise Exception("Cache file path cannot be None")

        self._logger.info("Writing the knowledge tree in cache '{}'".format(self.__cache_file_path))

        if os.path.exists(self.__cache_file_path):
            self._logger.info("Removing previous cache file '{}'".format(self.__cache_file_path))
            os.remove(self.__cache_file_path)

        nodes = [ root.serialize() for root in self.roots ]
        with open(self.__cache_file_path, "w") as fd:
            str_content = json.dumps(nodes, sort_keys=True, indent=4, separators=(',', ': '))
            fd.write(str_content)

    def load_cache(self, possible_letters):
        """This method loads the content of the cache in the knowledge tree

        See doctest declared in method "write_cache"
    
        """
        if self.__cache_file_path is None:
            raise Exception("Cache file path cannot be None")
        self._logger.info("Loading cache from '{}'".format(self.__cache_file_path))

        json_content = None

        with open(self.__cache_file_path, "r") as fd:
            json_content = json.loads(fd.read())

        for content in json_content:
            root = KnowledgeNode.deserialize(content, possible_letters)
            self.roots.append(root)
        
    def __add_letters(self, input_letters, output_letters):
        self._logger.debug("Adding letters '{}' / '{}'".format(', '.join([str(l) for l in input_letters]), ', '.join([str(l) for l in output_letters])))

        retained_root = None

        for root in self.roots:
            if root.input_letter == input_letters[0]:
                if root.output_letter != output_letters[0]:
                    raise Exception("Incompatible path found, expected '{}' found '{}'".format(root.output_letter.symbols, output_letters[0].symbols))
                retained_root = root
                break

        if retained_root is None:
            retained_root = KnowledgeNode(input_letters[0], output_letters[0])
            self._logger.debug("Creating '{}' as a new root".format(retained_root))
            self.roots.append(retained_root)

        return retained_root.traverse(input_letters, output_letters)



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------


# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger

@PylstarLogger
class Letter(object):
    """
    A letter is a wrapper for a set of symbols. A word is made of letters.
    A symbol can be of any type (e.g. str, int, Custom class)

    >>> from pylstar.Letter import Letter
    >>> l1 = Letter("l1")
    >>> print(l1)
    Letter('l1')
    
    """

    def __init__(self, symbol = None, symbols = None):
        self.symbols = set()
        
        if symbol is not None:
            self.symbols.add(symbol)
        if symbols is not None:
            self.symbols.update(symbols)

    def __hash__(self):
        return hash(frozenset(self.symbols))

    def __eq__(self, other):
        """Two letters are equal iif their symbols are equals

        >>> from pylstar.Letter import Letter
        >>> la = Letter("a")
        >>> lb = Letter("b")
        >>> la == lb
        False
        >>> la == "a"
        False
        >>> la == Letter("a")
        True
        """
        if not isinstance(other, Letter):
            return False
        
        return self.symbols == other.symbols

    def __ne__(self, other):
        """Two letters are not equal if their symbols are not equals

        >>> from pylstar.Letter import Letter
        >>> la = Letter("a")
        >>> lb = Letter("b")
        >>> la != lb
        True
        >>> la != "a"
        True
        >>> la != Letter("a")
        False
        """

        if not isinstance(other, Letter):
            return True
        return self.symbols != other.symbols
    
    def __str__(self):
        return "Letter({})".format(self.name)

    @property
    def name(self):                
        str_name = "None"
        if self.symbols is not None:
            str_name = ','.join([repr(s) for s in self.symbols])
        return str_name

    def __repr__(self):
        return self.__str__()

    def serialize(self):
        return self.name

    @staticmethod
    def deserialize(str_letters, possible_letters):
        letters = []
        for str_letter in str_letters.split(','):
            found = False
            for possible_letter in possible_letters:
                if possible_letter.name == str_letter:
                    letters.append(possible_letter)
                    found = True
                    break
            if not found:
                raise Exception("Cannot find any letter that fit with '{}'".format(str_letter))
        if len(letters) == 1:
            return letters[0]
        else:
            symbols = []
            for l in letters:
                symbols.extend(l.symbols)
            return Letter(symbols=symbols)

        
        
                


    @property
    def symbols(self):
        """Symbols that are represented by the letter"""
        return self.__symbols
    
    @symbols.setter
    def symbols(self, symbols):    
        self.__symbols = symbols

        
@PylstarLogger
class EmptyLetter(Letter):

    def __init__(self):
        super(EmptyLetter, self).__init__()

    def __str__(self):
        return "EmptyLetter"



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import uuid

# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.automata.Automata import Automata
from pylstar.automata.State import State
from pylstar.automata.Transition import Transition
from pylstar.Letter import Letter
from pylstar.tools.Decorators import PylstarLogger


@PylstarLogger
class DOTParser(object):
    """A parser for DOT files
    """

    @staticmethod
    def parse(dot_content, vocabulary = None):
        """This method returns the Automata that is represented by the specified DOT.
        
        :param dot_content: the DOT definition of the automata to build
        :type dot_content: str
        :rtype: pylstar.automata.Automata.Automata
        
        >>> from pylstar.automata.State import State
        >>> from pylstar.Letter import Letter
        >>> from pylstar.automata.Transition import Transition
        >>> from pylstar.automata.Automata import Automata
        >>> la = Letter('A')
        >>> lb = Letter('B')
        >>> l0 = Letter(0)
        >>> l1 = Letter(1)
        >>> q0 = State("Q0")
        >>> q1 = State("Q1")
        >>> q2 = State("Q2")
        >>> t0 = Transition("t0", q0, la, l0)
        >>> q0.transitions.append(t0)
        >>> t1 = Transition("t1", q1, lb, l1)
        >>> q0.transitions.append(t1)
        >>> t2 = Transition("t2", q0, la, l0)
        >>> q1.transitions.append(t2)
        >>> t3 = Transition("t3", q2, lb, l1)
        >>> q1.transitions.append(t3)
        >>> t4 = Transition("t4", q0, la, l1)
        >>> q2.transitions.append(t4)
        >>> automata = Automata(initial_state = q0, name = "Example")
        >>> dot_automata = DOTParser.build_dot_code(automata)
        >>> parsed_automata = DOTParser.parse(dot_automata)
        >>> print(parsed_automata.name)
        Example
        >>> print(DOTParser.build_dot_code(parsed_automata))
        digraph "Example" {
        "Q0" [shape=doubleoctagon, style=filled, fillcolor=white, URL="Q0"];
        "Q1" [shape=ellipse, style=filled, fillcolor=white, URL="Q1"];
        "Q2" [shape=ellipse, style=filled, fillcolor=white, URL="Q2"];
        "Q0" -> "Q0" [fontsize=5, label="A / 0", URL="t0"];
        "Q0" -> "Q1" [fontsize=5, label="B / 1", URL="t1"];
        "Q1" -> "Q0" [fontsize=5, label="A / 0", URL="t2"];
        "Q1" -> "Q2" [fontsize=5, label="B / 1", URL="t3"];
        "Q2" -> "Q0" [fontsize=5, label="A / 1", URL="t4"];
        }

        """

        
        if dot_content is None:
            raise Exception("dot_content cannot be None")
        dot_content = str(dot_content).strip()
        if len(dot_content) == 0:
            raise Exception("dot_content cannot be Empty")

        automata = None

        if not dot_content.startswith("digraph "):
            raise Exception("Dot_content must starts with digraph")

        i_start_graph_def = dot_content.find('{')
        if i_start_graph_def == -1:
            raise Exception("Cannot find '{' that indicates beginning of the graph definition")

        # extracts the automata name
        automata_name = dot_content[len("digraph "):i_start_graph_def].strip()
        if len(automata_name) == 0:
            raise Exception("Name of the automata cannot be None")

        automata_name = automata_name.replace('"', '')
        
        graph_def = dot_content[i_start_graph_def+1:].strip()
        graph_entries = graph_def.split(";")

        states = []
        # parses all the state definitions and their transitions
        for graph_entry in graph_entries:
            try:
                DOTParser.__parse_graph_entry(graph_entry, states)                    
            except Exception:
                pass

        automata = Automata(states[0], name=automata_name)

        return automata
    
    @staticmethod
    def __parse_graph_entry(graph_entry, states):
        if graph_entry is None:
            raise Exception("Graph entry cannot be None")
        graph_entry = str(graph_entry).strip()
        if len(graph_entry) == 0:
            raise Exception("Graph entry cannot be None")

        # parse first object
        i_start_first_obj = graph_entry.find('"')
        if i_start_first_obj == -1:
            raise Exception("Cannot find first object definition")

        i_end_first_obj = graph_entry.find('"', i_start_first_obj+1)
        if i_end_first_obj == -1:
            raise Exception("Cannot find first object definition")

        first_obj_name = graph_entry[i_start_first_obj+1:i_end_first_obj].strip()
        if len(first_obj_name) == 0:
            raise Exception("Cannot parse the name of the first object")

        first_state = None
        for state in states:
            if state.name == first_obj_name:
                first_state = state                

        if first_state is None:
            first_state = State(name=first_obj_name)
            states.append(first_state)

        remainder = graph_entry[i_end_first_obj+1:].strip()

        if remainder.startswith("->"):

            remainder = remainder[2:].strip()

            # parses the second state of the transition
            i_start_second_obj = remainder.find('"')
            if i_start_second_obj == -1:
                raise Exception("Cannot find second object definition")

            i_end_second_obj = remainder.find('"', i_start_second_obj+1)
            if i_end_second_obj == -1:
                raise Exception("Cannot find second object definition")

            second_obj_name = remainder[i_start_second_obj+1:i_end_second_obj].strip()
            if len(second_obj_name) == 0:
                raise Exception("Cannot find the name of the destination state")

            second_state = None
            for state in states:
                if state.name == second_obj_name:
                    second_state = state
                    
            if second_state is None:
                second_state = State(name=second_obj_name)
                states.append(second_state)

            # parses the transition input and output letters
            remainder = remainder[i_end_second_obj+2:]
            i_start_transition_details = remainder.find('[')
            if i_start_transition_details == -1:
                raise Exception("Cannot find transition details")
            
            transition_details = remainder[i_start_transition_details:]

            # parses the transition label
            i_start_label = transition_details.find('label=')
            if i_start_label == -1:
                raise Exception("Cannot find label of the transition")
            i_end_label = transition_details[i_start_label + len('label="'):].find('"')
            label = transition_details[i_start_label + len('label="'): i_start_label + len('label="')+i_end_label].strip()

            if len(label) == 0:
                raise Exception("Cannot find label")

            # parses input and output letters out of the label
            (input, output) = label.split('/')

            input_letter = Letter(input.strip())
            output_letter = Letter(output.strip())            

            # parses the transition name (url)
            i_start_url = transition_details.find('URL=')
            if i_start_url != -1:
                i_end_url = transition_details[i_start_url + len('URL="'):].find('"')
                url = transition_details[i_start_url + len('url="'): i_start_url + len('url="')+i_end_url].strip()
                t_name = url
            else:
                t_name = str(uuid.uuid4())

            transition = Transition(t_name, second_state, input_letter, output_letter)
            first_state.transitions.append(transition)    

    @staticmethod
    def build_dot_code(automata):
        """This method returns the DOT code that represents the provided Automata.

        :param automata: The automata the returned DOT code will represent
        :type automata: pylstar.automata.Automata
        :rtype: str        

        >>> from pylstar.automata.State import State
        >>> from pylstar.Letter import Letter
        >>> from pylstar.automata.Transition import Transition
        >>> from pylstar.automata.Automata import Automata
        >>> la = Letter('A')
        >>> lb = Letter('B')
        >>> l0 = Letter(0)
        >>> l1 = Letter(1)
        >>> q0 = State("Q0")
        >>> q1 = State("Q1")
        >>> q2 = State("Q2")
        >>> t0 = Transition("t0", q0, la, l0)
        >>> q0.transitions.append(t0)
        >>> t1 = Transition("t1", q1, lb, l1)
        >>> q0.transitions.append(t1)
        >>> t2 = Transition("t2", q1, la, l0)
        >>> q1.transitions.append(t2)
        >>> t3 = Transition("t3", q2, lb, l1)
        >>> q1.transitions.append(t3)
        >>> t4 = Transition("t4", q0, la, l1)
        >>> q2.transitions.append(t4)
        >>> automata = Automata(initial_state = q0, name = "Example")
        >>> print(DOTParser.build_dot_code(automata))
        digraph "Example" {
        "Q0" [shape=doubleoctagon, style=filled, fillcolor=white, URL="Q0"];
        "Q1" [shape=ellipse, style=filled, fillcolor=white, URL="Q1"];
        "Q2" [shape=ellipse, style=filled, fillcolor=white, URL="Q2"];
        "Q0" -> "Q0" [fontsize=5, label="A / 0", URL="t0"];
        "Q0" -> "Q1" [fontsize=5, label="B / 1", URL="t1"];
        "Q1" -> "Q1" [fontsize=5, label="A / 0", URL="t2"];
        "Q1" -> "Q2" [fontsize=5, label="B / 1", URL="t3"];
        "Q2" -> "Q0" [fontsize=5, label="A / 1", URL="t4"];
        }


        """

        if automata is None:
            raise Exception("Automata cannot be None")
        if not isinstance(automata, Automata):
            raise Exception("Must be an Automata")
        
        dot_code = []
        dot_code.append("digraph \"{}\" {{".format(automata.name))

        # Includes all the states declared in the automata
        states = automata.get_states()
        for state in states:
            color = "white"

            if state == automata.initial_state:
                shape = "doubleoctagon"
            else:
                shape = "ellipse"

            dot_code.append('"{0}" [shape={1}, style=filled, fillcolor={2}, URL="{3}"];'.format(state.name, shape, color, state.name))

        # Adds all the transitions
        for inputState in states:
            for transition in inputState.transitions:
                outputState = transition.output_state
                dot_code.append('"{0}" -> "{1}" [fontsize=5, label="{2}", URL="{3}"];'.format(inputState.name, outputState.name, transition.label, transition.name))

        dot_code.append("}")

        return '\n'.join(dot_code)




# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------


# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger


@PylstarLogger
class Transition(object):
    """Definition of a transition that belongs to an automata
    """

    def __init__(self, name, output_state, input_letter, output_letter):
        self.name = name
        self.output_state = output_state
        self.input_letter = input_letter
        self.output_letter = output_letter

    @property
    def label(self):
        input_symbols = ",".join([str(symbol) for symbol in self.input_letter.symbols])
        output_symbols = ",".join([str(symbol) for symbol in self.output_letter.symbols])        
        
        return "{} / {}".format(input_symbols, output_symbols)

        
        

        

    



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import itertools
from collections import deque

# +----------------------------------------------------------------------------
# | pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.Word import Word
from pylstar.Letter import EmptyLetter
from pylstar.OutputQuery import OutputQuery


@PylstarLogger
class WpMethodEQ(object):
    """WPmethod algorithm used to trigger an equivalence query"""

    def __init__(self, knowledge_base, max_states, input_letters):
        self.knowledge_base = knowledge_base
        self.max_states = max_states
        self.input_letters = input_letters

    def find_counterexample(self, hypothesis):
        if hypothesis is None:
            raise Exception("Hypothesis cannot be None")

        W = []

        states = hypothesis.get_states()
        
        # compute all couples of states
        state_couples = itertools.combinations(states, 2)

        # Constructing the characterization set W of the hypothesis
        for couple in state_couples:
            # Computes a distinguishing string for each couple of state
            W.append(self.__compute_distinguishable_string(hypothesis, couple))

        # computes P
        P = self.__computesP(hypothesis)
        self._logger.debug("P= {}".format(P))

        # computes Z
        Z = self.__computesZ(hypothesis, W)
        self._logger.debug("Z= {}".format(Z))

        # T = P . Z
        T =  P + Z
        self._logger.debug("T={}".format(T))

        # check if one of the computed testcase highlights a counterexample
        for i_testcase, testcase_query in enumerate(T[1:]):
            self._logger.debug("Executing testcase {}/{} : {}".format(i_testcase, len(T)-1, testcase_query))

            # computes the hypothesis output
            hypothesis_output_word = hypothesis.play_query(testcase_query)[0]

            self.knowledge_base.resolve_query(testcase_query)
            real_output_word = testcase_query.output_word

            self._logger.debug(real_output_word)
            self._logger.debug(hypothesis_output_word)
            if real_output_word != hypothesis_output_word:
                return testcase_query
        return None

    def __computesZ(self, hypothesis, W):    
        """it follows the formula Z= W U (X^1.W) U .... U (X^(m-1-n).W) U (W^(m-n).W)

        """
        if hypothesis is None:
            raise Exception("Hypothesis cannot be None")
        if W is None:
            raise Exception("W cannot be None")
            
        self._logger.debug("Computing Z")

        Z = []
        Z.extend(W)
        
        states = hypothesis.get_states()
        v = self.max_states - len(states)
        if v < 0:
            v = 0
        self._logger.debug("V= {}".format(v))

        output_queries = []
        for input_letter in self.input_letters:
            output_query = OutputQuery(word = Word([input_letter]))
            output_queries.append(output_query)

        X = dict()
        X[0] = W
        for i in range(1, v+1):
            self._logger.debug("Computing X^{}".format(i))
            X[i] = []
            previous_X = X[i-1]
            for x in previous_X:
                X[i].extend(x.multiply(output_queries))
            for w in W:
                for xi in X[i]:
                    if not xi in Z:
                        Z.append(xi)

        return Z

    def __computesP(self, hypothesis):
        if hypothesis is None:
            raise Exception("Hypothesis cannot be None")
        self._logger.debug("Computing P")

        P = []
            
        empty_word = Word([EmptyLetter()])
        current_query = OutputQuery(empty_word)
        P.append(current_query)

        open_queries = deque([current_query])
        close_queries = []

        seen_states = set([hypothesis.initial_state])
        while len(open_queries) > 0:
            query = open_queries.popleft()
            tmp_seen_states = set()

            for letter in self.input_letters:
                new_word = query.input_word + Word([letter])
                query_z = OutputQuery(new_word)
                (output_word, visited_states) = hypothesis.play_query(query_z)
                close_queries.append(query_z)
                
                if visited_states[-1] not in seen_states:
                    tmp_seen_states.add(visited_states[-1])
                    open_queries.append(query_z)

            seen_states.update(tmp_seen_states)

        P.extend(close_queries)

        return P
        
    def __compute_distinguishable_string(self, hypothesis, couple):
        self._logger.debug("Computes the distinguishable string for state couple '{}'".format(couple))
        if hypothesis is None:
            raise Exception("Hypothesis cannot be None")
        if couple is None:
            raise Exception("couple cannot be None")
            
        self._logger.debug("Computing distinguishing strings for states {}".format(couple))
        queries_to_test = deque([])
        
        empty_word = Word([EmptyLetter()])
        z_query = OutputQuery(empty_word)
        for letter in self.input_letters:
            new_word = z_query.input_word + Word([letter])
            queries_to_test.append(OutputQuery(new_word))

        distinguishable_query = z_query

        done = False
        i = 0
        while not done:
            query = queries_to_test.popleft()
            if i > self.max_states * self.max_states:
                break

            if not self.__is_distinguishable_states(hypothesis, query, couple):
                done = False
                for letter in self.input_letters:
                    new_query = OutputQuery(query.input_word + Word([letter]))
                    queries_to_test.append(new_query)
            else:
                done = True
                distinguishable_query = query

            i = i + 1

        return distinguishable_query
        
    def __is_distinguishable_states(self, hypothesis, query, couple):
        if hypothesis is None:
            raise Exception("Hypothesis cannot be None")
        if query is None:
            raise Exception("query cannot be None")
        if couple is None:
            raise Exception("couple cannot be None")

        output_word_state0 = hypothesis.play_word(query.input_word, couple[0])[0]
        output_word_state1 = hypothesis.play_word(query.input_word, couple[1])[0]
        
        return output_word_state0 != output_word_state1




# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import abc

# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger


@PylstarLogger
class KnowledgeBaseStats(object):

    def __init__(self):
        self.nb_query = 0
        self.nb_submited_query = 0
        self.nb_letter = 0
        self.nb_submited_letter = 0
        
    def __str__(self):
        return """
\t- nb query= {nb_query}
\t- nb submited query= {nb_submited_query}
\t- nb letter= {nb_letter}
\t- nb submited letter= {nb_submited_letter}

""".format(nb_query = self.nb_query,
               nb_letter = self.nb_letter,
               nb_submited_query = self.nb_submited_query,
               nb_submited_letter = self.nb_submited_letter)

    @property
    def nb_submited_query(self):
        """Number of query submited to the target"""
        return self.__nb_submited_query

    @nb_submited_query.setter
    def nb_submited_query(self, nb_submited_query):
        if nb_submited_query is None:
            raise Exception("Nb submited query cannot be None")
        nb_submited_query = int(nb_submited_query)
        if nb_submited_query < 0:
            raise Exception("Nb submited query must be > 0")
        
        self.__nb_submited_query = nb_submited_query

    @property
    def nb_submited_letter(self):
        """Number of letter submited to the target"""
        return self.__nb_submited_letter

    @nb_submited_letter.setter
    def nb_submited_letter(self, nb_submited_letter):
        if nb_submited_letter is None:
            raise Exception("Nb submited letter cannot be None")
        nb_submited_letter = int(nb_submited_letter)
        if nb_submited_letter < 0:
            raise Exception("Nb submited letter must be > 0")
        
        self.__nb_submited_letter = nb_submited_letter

    
    @property
    def nb_letter(self):
        """Number of letters triggered while infering"""
        return self.__nb_letter

    @nb_letter.setter
    def nb_letter(self, nb_letter):
        if nb_letter is None:
            raise Exception("Nb letter cannot be None")
        nb_letter = int(nb_letter)
        if nb_letter < 0:
            raise Exception("Nb letter must be > 0")
        
        self.__nb_letter = nb_letter

    @property
    def nb_query(self):
        """Number of queries triggered while infering"""
        return self.__nb_query

    @nb_query.setter
    def nb_query(self, nb_query):
        if nb_query is None:
            raise Exception("Nb query cannot be None")
        nb_query = int(nb_query)
        if nb_query < 0:
            raise Exception("Nb query must be > 0")
        
        self.__nb_query = nb_query
        
        

    



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------


# +----------------------------------------------------------------------------
# | Pylstar Imports
# +----------------------------------------------------------------------------
from pylstar.tools.Decorators import PylstarLogger
from pylstar.Word import Word
from pylstar.automata.State import State


@PylstarLogger
class Automata(object):
    """Definition of an automata
    """

    def __init__(self, initial_state, name = "Automata"):
        self.initial_state = initial_state
        self.name = name

    def play_query(self, query):
        return self.play_word(query.input_word, self.initial_state)
        
                   
        
    def play_word(self, input_word, starting_state=None):
        """This method can be used to play the specified word
        accross the current automata.
        It returns a tupple made of the output_word and the visited state
        captured while visiting the automata

        >>> from pylstar.Letter import Letter, EmptyLetter
        >>> from pylstar.Word import Word
        >>> from pylstar.automata.State import State
        >>> from pylstar.automata.Transition import Transition
        >>> from pylstar.automata.Automata import Automata
        >>> l_lambda = EmptyLetter()
        >>> l_a = Letter('a')
        >>> l_b = Letter('b')
        >>> l_0 = Letter(0)
        >>> l_1 = Letter(1)
        >>> s0 = State("S0")
        >>> s1 = State("S1")
        >>> s2 = State("S2")
        >>> s3 = State("S3")
        >>> t1 = Transition("T1", s3, l_a, l_0)
        >>> t2 = Transition("T2", s1, l_b, l_0)        
        >>> s0.transitions = [t1, t2]
        >>> t3 = Transition("T3", s0, l_a, l_1)
        >>> t4 = Transition("T4", s2, l_b, l_1)        
        >>> s1.transitions = [t3, t4]
        >>> t5 = Transition("T5", s3, l_a, l_0)
        >>> t6 = Transition("T6", s0, l_b, l_0)        
        >>> s2.transitions = [t5, t6]        
        >>> t7 = Transition("T7", s3, l_a, l_1)
        >>> t8 = Transition("T8", s3, l_b, l_1)        
        >>> s3.transitions = [t7, t8]
        >>> automata = Automata(s0)
        >>> print(automata.play_word(Word([l_a, l_a, l_a]))[0])
        [Letter(0), Letter(1), Letter(1)]
        >>> print(automata.play_word(Word([l_b, l_b, l_b]))[0])
        [Letter(0), Letter(1), Letter(0)]
        >>> print(automata.play_word(Word([l_b, l_a,  l_b, l_a, l_b]))[0])
        [Letter(0), Letter(1), Letter(0), Letter(1), Letter(0)]
        
        """

        if input_word is None or len(input_word) == 0:
            raise Exception("Input word cannot be None or empty")

        if starting_state is None:
            current_state = self.initial_state
        else:
            current_state = starting_state

        self._logger.debug("Playing word '{}'".format(input_word))

        output_letters = []
        visited_states = []
        
        for letter in input_word.letters:
            (output_letter, output_state) = current_state.visit(letter)
            output_letters.append(output_letter)
            visited_states.append(output_state)

            current_state = output_state

        output_word = Word(letters=output_letters)
        return (output_word, visited_states)        


    def get_states(self):
        """Visits the automata to discover all the available states.

        :return: a list containing all the discovered states.
        :rtype: a :class:`list`
        """

        states = []
        toAnalyze = []
        toAnalyze.append(self.initial_state)
        while (len(toAnalyze) > 0):
            currentState = toAnalyze.pop()
            if currentState is not None:
                found = False
                for tmpState in states:
                        if tmpState.name == currentState.name:
                            found = True
                if not found:
                    for transition in currentState.transitions:
                        outputState = transition.output_state
                        found = False
                        for tmpState in states:
                            if tmpState.name == outputState.name:
                                found = True
                        for tmpState in toAnalyze:
                            if tmpState.name == outputState.name:
                                found = True
                        if not found:
                            toAnalyze.append(outputState)
                    states.append(currentState)
        return states

    @staticmethod
    def create_from_dot_code(dot_code):
        """This statis method returns the Automata object that can represents the provided DOT code

        :param dot_code: DOT definition of the Automata to parse
        :type dot_code: str
        :rtype: pylstar.automata.Automata.Automata
        """

        if dot_code is None:
            raise Exception("Dot code cannot be None")

        if not isinstance(dot_code, str):
            raise Exception("Dot code must be a String")

        from pylstar.automata.DOTParser import DOTParser

        return DOTParser.parse(dot_code)
        
        

    def build_dot_code(self):
        """Returns the dot code representing the automata.

        :rtype: str        

        >>> from pylstar.automata.State import State
        >>> from pylstar.Letter import Letter
        >>> from pylstar.automata.Transition import Transition
        >>> la = Letter('A')
        >>> lb = Letter('B')
        >>> l0 = Letter(0)
        >>> l1 = Letter(1)
        >>> q0 = State("Q0")
        >>> q1 = State("Q1")
        >>> t0 = Transition("t0", q0, la, l0)
        >>> q0.transitions.append(t0)
        >>> t1 = Transition("t1", q1, lb, l1)
        >>> q0.transitions.append(t1)
        >>> t2 = Transition("t2", q1, la, l0)
        >>> q1.transitions.append(t2)
        >>> t3 = Transition("t3", q0, lb, l1)
        >>> q1.transitions.append(t3)
        >>> automata = Automata(initial_state = q0)
        >>> print(automata.build_dot_code())
        digraph "Automata" {
        "Q0" [shape=doubleoctagon, style=filled, fillcolor=white, URL="Q0"];
        "Q1" [shape=ellipse, style=filled, fillcolor=white, URL="Q1"];
        "Q0" -> "Q0" [fontsize=5, label="A / 0", URL="t0"];
        "Q0" -> "Q1" [fontsize=5, label="B / 1", URL="t1"];
        "Q1" -> "Q1" [fontsize=5, label="A / 0", URL="t2"];
        "Q1" -> "Q0" [fontsize=5, label="B / 1", URL="t3"];
        }


        """
        from pylstar.automata.DOTParser import DOTParser

        return DOTParser.build_dot_code(self)

        
    @property
    def initial_state(self):
        """The initial state of the Automata"""
        return self.__initial_state

    @initial_state.setter
    def initial_state(self, state):
        if state is None:
            raise Exception("Initial state cannot be None")
        if not isinstance(state, State):
            raise Exception("Initial state must be a state")
        self.__initial_state = state

    @property
    def name(self):
        """The name of the state machine"""
        return self.__name

    @name.setter
    def name(self, name):
        if name is None:
            raise Exception("Name of the automata cannot be None")
        if not isinstance(name, str):
            raise Exception("Name of the automata must be a String")
        self.__name = name
        

    



# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

#+---------------------------------------------------------------------------+
#| Standard library imports                                                  |
#+---------------------------------------------------------------------------+
import logging
logging.basicConfig(level=logging.DEBUG)

#+---------------------------------------------------------------------------+
#| Related third party imports                                               |
#+---------------------------------------------------------------------------+
from functools import wraps

#+---------------------------------------------------------------------------+
#| Local application imports                                                 |
#+---------------------------------------------------------------------------+

# Definition of the ColorStreamHandler class only if dependency colorama is
# available on the current system.
try:
    from colorama import Fore, Back, Style

    class ColourStreamHandler(logging.StreamHandler):
        """ A colorized output SteamHandler """

        # Some basic colour scheme defaults
        colours = {
            'DEBUG': Fore.CYAN,
            'INFO': Fore.GREEN,
            'WARN': Fore.YELLOW,
            'WARNING': Fore.YELLOW,
            'ERROR': Fore.RED,
            'CRIT': Back.RED + Fore.WHITE,
            'CRITICAL': Back.RED + Fore.WHITE
        }

        @property
        def is_tty(self):
            """ Check if we are using a "real" TTY. If we are not using a TTY it means that
            the colour output should be disabled.

            :return: Using a TTY status
            :rtype: bool
            """
            try:
                return getattr(self.stream, 'isatty', None)()
            except:
                return False

        def emit(self, record):
            try:
                message = self.format(record)

                if not self.is_tty:
                    self.stream.write(message)
                else:
                    self.stream.write(self.colours[record.levelname] + message + Style.RESET_ALL)
                self.stream.write(getattr(self, 'terminator', '\n'))
                self.flush()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                self.handleError(record)

    has_colour = True
except Exception as e:
    has_colour = False


def PylstarLogger(klass):
    """This class decorator adds (if necessary) an instance
    of the logger (self.__logger) to the attached class
    and removes from the getState the logger.

    """

    # Verify if a logger already exists
    found = False
    for k, v in klass.__dict__.items():
        if isinstance(v, logging.Logger):
            found = True
            break
    if not found:
        klass._logger = logging.getLogger(klass.__name__)
        handler = ColourStreamHandler() if has_colour else logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        fmt = '%(relativeCreated)d: [%(levelname)s] %(module)s:%(funcName)s: %(message)s'
        handler.setFormatter(logging.Formatter(fmt))
        klass._logger.addHandler(handler)
        klass._logger.propagate = False

    # Exclude logger from __getstate__
    def getState(self, **kwargs):
        r = dict()
        for k, v in self.__dict__.items():
            if not isinstance(v, logging.Logger):
                r[k] = v
        return r

    def setState(self, dict):
        self.__dict__ = dict
        self.__logger = logging.getLogger(klass.__name__)

    klass.__getstate__ = getState
    klass.__setState__ = setState

    return klass


def typeCheck(*types):
    """Decorator which reduces the amount of code to type-check attributes.

    Its allows to replace the following code:
    ::
        @id.setter
        def id(self, id):
            if not isinstance(id, uuid.UUID):
               raise TypeError("Invalid types for argument id, must be an UUID")
            self.__id = id

    with:
    ::
        @id.setter
        @typeCheck(uuid.UUID)
        def id(self, id):
           self.__id = id

    .. note:: set type = "SELF" to check the type of the self parameter
    .. warning:: if argument is None, the type checking is not executed on it.

    """
    def _typeCheck_(func):
        def wrapped_f(*args, **kwargs):
            arguments = args[1:]
            if len(arguments) == len(types):
                # Replace "SELF" with args[0] type
                final_types = []
                for type in types:
                    if type == "SELF":
                        final_types.append(args[0].__class__)
                    else:
                        final_types.append(type)

                for i, argument in enumerate(arguments):
                    if argument is not None and not isinstance(argument, final_types[i]):
                        raise TypeError("Invalid type for arguments, expecting: {0} and received {1}".format(', '.join([t.__name__ for t in final_types]), argument.__class__.__name__))
            return func(*args, **kwargs)
        return wraps(func)(wrapped_f)
    return _typeCheck_



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# +---------------------------------------------------------------------------+
# | pylstar : Implementation of the LSTAR Grammatical Inference Algorithm     |
# +---------------------------------------------------------------------------+
# | Copyright (C) 2015 Georges Bossert                                        |
# | This program is free software: you can redistribute it and/or modify      |
# | it under the terms of the GNU General Public License as published by      |
# | the Free Software Foundation, either version 3 of the License, or         |
# | (at your option) any later version.                                       |
# |                                                                           |
# | This program is distributed in the hope that it will be useful,           |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              |
# | GNU General Public License for more details.                              |
# |                                                                           |
# | You should have received a copy of the GNU General Public License         |
# | along with this program. If not, see <http://www.gnu.org/licenses/>.      |
# +---------------------------------------------------------------------------+
# | @url      : https://github.com/gbossert/pylstar                           |
# | @contact  : gbossert@miskin.fr                                            |
# +---------------------------------------------------------------------------+

# +----------------------------------------------------------------------------
# | Global Imports
# +----------------------------------------------------------------------------
import sys

from setuptools import setup, find_packages

sys.path.insert(0, 'src/')
from pylstar import release
from resources.sdist.test_command import test_command
from resources.sdist.utils import opj

# +----------------------------------------------------------------------------
# | Definition of the dependencies
# +----------------------------------------------------------------------------
dependencies = []
with open('requirements.txt', 'r') as fd_requirements:
    for dependency in fd_requirements:
        dependencies.append(dependency.strip())

extra_dependencies = {
    'docs': ['Sphinx>=1.1.3']
}

dependency_links = []

data_files = []

# Extract the long description from README.rst and NEWS.rst files
README = open('README.rst', 'rt').read()
NEWS = open('CHANGELOG.rst', 'rt').read()

# +----------------------------------------------------------------------------
# | Extensions in the build operations (test, ...)
# +----------------------------------------------------------------------------
CMD_CLASS = {
    'test': test_command
}
# +----------------------------------------------------------------------------
# | Definition of the package
# +----------------------------------------------------------------------------
setup(
    name=release.name,
    packages=find_packages(where='src'),
    package_dir={
        "pylstar": opj("src", "pylstar")
    },
    data_files=data_files,
    install_requires=dependencies,
    extras_require=extra_dependencies,
    dependency_links=dependency_links,
    version=release.version,
    license=release.licenseName,
    description=release.description,
    platforms=release.platforms,
    author=release.author,
    author_email=release.author_email,
    url=release.url,
    download_url=release.download_url,
    keywords=release.keywords,
    long_description=README + '\n' + NEWS,
    cmdclass=CMD_CLASS,
)

